{--------------------------------------------------------------------------------}
{                                                                                }
{                            UNIT : FLEXIREP                                     }
{                                                                                }
{   TITLE       : Flexi Report                                                   }
{                                                                                }
{   EXPLANATION : A class which encapsulates the body of a report,               }
{                 allowing printing and exporting.                               }
{                 Used in conjunction with the report generator, and report      }
{                 charting classes.                                              }
{                                                                                }
{   DATE        : 22/04/97                                                       }
{   AUTHOR      : P.CHAPMAN                                                      }
{   STATUS      : Development                                                    }
{                                                                                }
{--------------------------------------------------------------------------------}
{  Modifications                                                                 }
{                                                                                }
{  W2 02/06/97.PAC - Added ShowExportOptionsDialog routine to allow object to    }
{                    invoke its own dialog boxes                                 }          
{--------------------------------------------------------------------------------}
unit Flexirep;

interface

uses
  Classes,
  Controls,
  Grids,
  PRN_Type,
  PrintRep,
  WinTypes,
  Dialogs,
  XL_Types;

type
  { simple integer wrapper }
  TColumnWidth = class
  private
    Fwidth :integer;    { the length of the longest string in the column }
  public
    property Width :integer     read FWidth        write FWidth;
  end;

  TFlexiReport = class(TWinControl)
  private
    FGrids :TList;
    { collection of grids which hold the body of the report }

    FPrintedReport :TPrintedReport;
    { Holds the printed report object. Most access to this object is through the }
    { two structures passed in the print method of the flexi report, and through the }
    { various events it has to offer, although it is available as a property for }
    { changes which may be required inside its events }

    FExcelSheet :TExcelSheet;

    FLocalCollection :Boolean;
    { this is set if the FGrids collection was created locally, not passed in as a parameter }

    FConfirmDialog :Boolean;
    { is the flexireport allowed to create its own dialog boxes ? }

    FOnExportCell :PRN_TExportCellEvent;
    FOnExportGetColWidth :PRN_TExportGetColWidthEvent;

    FExportInfo :PRN_TExportInfoStruct;

    { PROPERTY ACCESS METHODS }
    procedure setPrintInfo(aPrintInfo :PRN_TPrintInfoStruct);
    function getPrintInfo :PRN_TPrintInfoStruct;
    procedure ShowConfirmDialog(AOwner :TComponent; Filename :string; ExportType :PRN_TExportType);

  public

    constructor Create(aOwner :TComponent; aGrid :TStringGrid);
    constructor CreateMulti(aOwner :TComponent; aGridCollection :TList);
    destructor Destroy; override;

  {---------}
  { METHODS }
  {---------}

    procedure Print(PrintInfo :PRN_TPrintInfoStruct; DefaultDrawInfo :PRN_TDrawInfoStruct);
    { prints the current collection of grids }
    { requires filled records on how to print, and how to draw each cell }

    procedure ShowExportOptionsDialog(Sender: TObject);
    { prompts the user for export options, and presents the save dialog }

    procedure ExportReportToFixedColumn(Filename :string);
    { exports the report to fixed column text file }

    procedure ExportReportWithDelimiter(Filename :string; delim :string);
    { exports the report to a CSV or similar delimited text file }

    procedure ExportReportToExcel(Filename :string);
    { exports the report to an Excel worksheet }

    procedure SetSaveDialogFilters(aSaveDialog :TSaveDialog; ExportType :PRN_TExportType);
    { given a savedialog, and a type of export, sets the save dialogs filters accordingly }

    procedure Initialise;
    { initialisation routine. Called by both constructors }

  {------------}
  { PROPERTIES }
  {------------}

    property ExcelSheet :TExcelSheet              read FExcelSheet      write FExcelSheet;
    { access to the underlying excel work sheet object }

    property PrintedReport :TPrintedReport        read FPrintedReport   write FPrintedReport;
    { access to the underlying printed report }

    property Grids :TList                         read FGrids           write FGrids;
    { provide access to the grids which were passed in through the constructor }

    property PrintInfo :PRN_TPrintInfoStruct      read getPrintInfo     write SetPrintInfo;
    { structure which defines the characteristics of the output of the printed report }

    property ExportInfo :PRN_TExportInfoStruct    read FExportInfo      write FExportInfo;
    { structure which defined the attributes of the current cell being exported }

    property OnExportCell :PRN_TExportCellEvent   read FOnExportCell    write FOnExportCell;
    { event handler which is called just before a cell is exported, for all three types of export }

    property OnExportGetColWidth :PRN_TExportGetColWidthEvent
          read FOnExportGetColWidth   write FOnExportGetColWidth;
    { event handler which is triggered for each column when exporting to fixed column text file }
    { this allows the user to specifically set or validate each columns width }

    property ConfirmDialog :Boolean               read FConfirmDialog   write FConfirmDialog;
    { flag to determine whether the flexireport is allowed to create its own dialogs }

  end;

implementation

uses
  SysUtils,
  Rep_Opts,
  Forms,
  GP_Util,
  ExpCnDlg;

{--------------}
{ CONSTRUCTORS }
{--------------}
constructor TFlexiReport.Create(aOwner :TComponent; aGrid :TStringGrid);
begin
  FGrids := nil;
  FPrintedReport := nil;

  inherited create(aOwner);

  FGrids := TList.create;
  FGrids.Add(aGrid);
  FLocalCollection := True;
  FOnExportCell := nil;
  FConfirmDialog := True;
  Initialise;
end;

constructor TFlexiReport.CreateMulti(aOwner :TComponent; aGridCollection :TList);
begin
  FGrids := nil;
  FPrintedReport := nil;

  inherited create(aOwner);

  FGrids := aGridCollection;
  FLocalCollection := False;
  Initialise;
end;

procedure TFlexiReport.Initialise;
begin
  OnExportGetColWidth := nil;
  OnExportCell := nil;

  FConfirmDialog := True;
end;

{------------}
{ DESTRUCTOR }
{------------}
destructor TFlexiReport.Destroy;
begin
  if (FLocalCollection and assigned(FGrids)) then
  begin
    FGrids.Clear;
    FGrids.Free;
  end;

  inherited Destroy;
end;

{-------------------------}
{ PROPERTY ACCESS METHODS }
{-------------------------}
procedure TFlexiReport.setPrintInfo(aPrintInfo :PRN_TPrintInfoStruct);
begin
  if (assigned(FPrintedReport)) then
    FPrintedReport.PrintInfo := aPrintinfo;
end;

function TFlexiReport.getPrintInfo :PRN_TPrintInfoStruct;
begin
  if (assigned(FPrintedReport)) then
    result := FPrintedReport.PrintInfo;
end;


{--------------}
{ PRINT REPORT }
{--------------}
procedure TFlexiReport.Print(PrintInfo :PRN_TPrintInfoStruct; DefaultDrawInfo :PRN_TDrawInfoStruct);
begin
  { create a printed report }
  FPrintedReport := TPrintedReport.create(self, FGrids);

  { set the structure properties }
  FPrintedReport.PrintInfo := PrintInfo;
  FPrintedReport.DefaultDrawInfo := DefaultDrawInfo;

  try
    { print the report }
    FPrintedReport.Print;
  finally
    { destroy the report }
    FPrintedReport.Free;
    FPrintedReport := nil;
  end;
end;

{----------------------------}
{ SHOW EXPORT OPTIONS DIALOG }
{----------------------------}
procedure TFlexiReport.ShowExportOptionsDialog(Sender: TObject);
var
  delimchar :string[1];
  SaveDialog1 :TSaveDialog;
begin
  try
    if not assigned(ExportOptionsDialog) then
      ExportOptionsDialog := TExportOptionsDialog.create(self, dcExportOptions);

    { if the exporting options dialog box was Okayed }
    if (ExportOptionsDialog.ShowModal = mrOK) then
    begin

      { create the save dialog to prompt the user for the filename }
      SaveDialog1 := TSaveDialog.create(Application);

      { set the filters of the dialog according to the export type selected }
      SetSaveDialogFilters(SaveDialog1, ExportOptionsDialog.ExportType);

      { if the save dialog box is not cancelled }
      { then export the report according to the export type requested }
      if (SaveDialog1.execute) then
      begin
        { excel 4 worksheet }
        if (ExportOptionsDialog.ExportType = etExcel) then
          ExportReportToExcel(SaveDialog1.Filename)

        else
        { delimited text file }
        if (ExportOptionsDialog.ExportType = etDelimited) then
        begin
          DelimChar := copy(ExportOptionsDialog.DelimiterEditBox.text, 1, 1);
          ExportReportWithDelimiter(SaveDialog1.Filename, DelimChar) { delimited }
        end

        else
        { fixed column text }
        if (ExportOptionsDialog.ExportType = etFixed) then
          ExportReportToFixedColumn(SaveDialog1.Filename)
      end;
    end;
  finally
    { no matter what, destroy the export options dialog }
    ExportOptionsDialog.Free;
    ExportOptionsDialog := nil;
  end;
end;

{---------------------------}
{ EXPORT TO EXCEL WORKSHEET }
{---------------------------}
procedure TFlexiReport.ExportReportToExcel(Filename :string);
var
  CurrentGrid :TStringGrid;
  iGrid, Row, Col :Integer;
  success :Boolean;
  uniRow, uniCol :integer;

     procedure ProcessCell(aCellStr :string; aCellType :PRN_TCellType);
     var
       tmpdouble :double;
     begin
       { check the publish state and add an appropriate cell }
       case aCellType of
         ctString: FExcelSheet.AddText(unirow, unicol, aCellStr);
         ctReal:
         begin
           tmpdouble := GP_Str2Real(aCellStr);
           FExcelSheet.AddNumber(unirow, unicol, tmpdouble);
         end;
       end;
     end;

begin
  unicol := 0;
  unirow := 0;
  success := True;

  try
    try
      { create the XL-worksheet writer object }
      FExcelSheet := TExcelSheet.Create;
      FExcelSheet.FileName := FileName;

     { for each grid in the collection }
      for iGrid := 0 to FGrids.count - 1 do
      begin
        CurrentGrid := FGrids.Items[iGrid];
        for Row := 0 to CurrentGrid.RowCount - 1 do
        begin
          UniCol := 0;

          for Col := 0 to CurrentGrid.ColCount - 1 do
          begin
            { set the fields required by the event handler }

            { populate the export info structure }
            FExportInfo.iRow := Row;
            FExportInfo.iCol := Col;
            with FExportInfo do
            begin
              Grid := CurrentGrid;
              CellText := CurrentGrid.Cells[Col, Row];
              TextJustify := DT_LEFT;
              CellType := ctString;
              ExportType := etExcel;
            end;

            { trigger the OnExport event handler if required }
            if assigned(FOnExportCell) then
            begin
              { trigger the event, returning the revised celltext and textjustify }
              FOnExportCell(self, FExportInfo);
            end;

            { export the cell using the revised content and type }
            ProcessCell(FExportInfo.CellText, FExportInfo.CellType);
            inc(unicol);
          end; { for each col }
          inc(unirow);
        end; { for each row }
      end; { for each grid }

     { output the excelsheet }
      FExcelSheet.WriteToFile;
    except
      success := False;

      { show an error dialog, if the report is allowed to issue its own dialogs }
      if (ConfirmDialog) then
        MessageDLG('Error: There was an error trying to save to the file: ' + ExtractFileName(FileName) +
                   '.' + #13#13 +
                    'Solution: Ensure that the file is not in use, is writable, and' +
                    ' the filename is valid.', mtError, [mbOK], 0);
    end;
  finally
     { free the excel sheet }
      FExcelSheet.Free;
      FExcelSheet := nil;

     { restore the cursor }
//      GP_RestoreCursor(Self);

     { show the success message if report is allowed to issue its own dialogs }
      if (success and ConfirmDialog) then
        ShowConfirmDialog(self, filename, etExcel);
  end;
end;

{------------------------}
{ EXPORT TO FIXED COLUMN }
{------------------------}
procedure TFlexiReport.ExportReportToFixedColumn(filename :string);
var
  iGrid,
  Row,
  Col :integer;

  CurrentGrid :TStringGrid;
  outfile :TFileStream;
{  CellType :PRN_TCellType; }

  success :boolean;
  widths :TList;
  tempColWidth :TColumnWidth;
  pipechar :string;
  CellStr :string;
  CellLength :integer;
  charno :integer;
  tmpwidth :integer;
  MaxColumns :integer;
  CurrentColumn :TColumnWidth;
const
  CarriageReturn :string[2] = #13#10;
begin
  outFile := nil;
  success := True;
  pipechar := '|';

  { >> we need to store the length of the longest string in each column along }
  { with its justification find the grid with the most columns. This is so we }
  { know the max no of column objects are required }
  MaxColumns := 0;
  for iGrid := 0 to FGrids.Count - 1 do
    MaxColumns := GP_Max(MaxColumns, TStringGrid(FGrids.Items[iGrid]).ColCount);

  { create the collection }
  Widths := TList.create;

  { create <MaxColumns> column objects (one for each col in the widest grid) }
  for Col := 0 to MaxColumns - 1 do
  begin
    { create and store add the ColumnWidth placeholder object }
    tempColWidth := TColumnWidth.create;
    Widths.Add(tempColWidth);
  end;

  { for each grid in the collection }
  for iGrid := 0 to FGrids.count - 1 do
  begin
    CurrentGrid := TStringGrid(FGrids.Items[iGrid]);

    { for each column in the current grid }
    for Col := 0 to CurrentGrid.ColCount - 1 do
    begin
      CurrentColumn := TColumnWidth(Widths.Items[Col]);

      { for each row in the grid }
      for Row := 0 to Currentgrid.RowCount - 1 do
      begin
        { get the string in the current row/col and its length }
        CellStr := CurrentGrid.Cells[Col, Row];
        CellLength := Length(CellStr);

        { if its the longest string found so far }
        { then update the ColumnWidth instance with the revised maximum length }
        if (CellLength > CurrentColumn.Width) then
          CurrentColumn.Width := CellLength;
      end;
    end;
  end;

  { trigger the OnGetFixedColumnWidth event handler if required }
  { this allows the user to revise the fixed column width, before it is committed }
  if assigned(FOnExportGetColWidth) then
    for Col := 0 to MaxColumns - 1 do
    begin
      CurrentColumn := TColumnWidth(Widths.Items[Col]);
      tmpWidth := CurrentColumn.Width;
      FOnExportGetColWidth(self, Col, tmpWidth);

      { commit the returned column width }
      CurrentColumn.Width := tmpWidth;
    end;

  try
    try
//      GP_WaitCursor(self);
      outfile := TFileStream.create(Filename, fmCreate);

      for iGrid := 0 to FGrids.count - 1 do
      begin
        { set a pointer to the currently processing grid }
        CurrentGrid := TStringGrid(FGrids.Items[igrid]);

        for Row := 0 to Currentgrid.RowCount - 1 do
        begin
          for Col := 0 to CurrentGrid.ColCount - 1 do
          begin
            { populate the ExportInfo structure }
            with FExportInfo do
            begin
              Grid := CurrentGrid;
              CellText :=  CurrentGrid.Cells[Col, Row];
              iRow := Row;
              iCol := Col;
              TextJustify := DT_LEFT;
              CellType := ctString;
              ExportType := etFixed;
            end;

            { trigger the OnExport event handler if required }
            if assigned(FOnExportCell) then
            begin
              { trigger the event, returning the revised celltext and textjustify }
              FOnExportCell(self, FExportInfo);
            end;

            { get the max string length recorded earlier, for the current column }
            CellLength := TColumnWidth(Widths.Items[Col]).width;

            { if the celltext is shorter than the column width, then pad it according to justification }
            while (Length(FExportInfo.CellText) < CellLength) do
            begin
              case FExportInfo.TextJustify of
                DT_LEFT:   FExportInfo.CellText := FExportInfo.CellText + ' ';
                DT_RIGHT:  FExportInfo.CellText := ' ' + FExportInfo.CellText;
              end;
            end;

            { write the string out to the stream, excluding the byte for strings size }
            for charno := 1 to Length(FExportInfo.CellText) do
              outfile.write(FExportInfo.CellText[charno], 1);

            { write the column separating character }
            outfile.write(PipeChar[1], 1);
          end;

          { we have completed a row of the grid, so end the line in the text file }
          outfile.write(CarriageReturn[1], 1);
          outfile.write(CarriageReturn[2], 1);
        end;
      end;

    except
      { if we get a file create error, and the report is allowed to issue its own dialogs }
      on EFCreateError do
      begin
        { then inform the user of the file creation error }
        if (FConfirmDialog) then
          MessageDLG('Error: There was an error trying to save to the specified file.' + #13 +
                     'This may have been caused by a disk/file error.'+ #13#13 +
                     'Solution: Ensure that the file is not in use, is writable and' +
                     ' the filename is valid.', mtError, [mbOK], 0);
        success := False;
      end;
    end;

  finally
    { close the stream cleanly, no matter what }
    if (outFile <> nil) then
      outfile.Free;

    { destroy the collection of column widths }
    Widths.Free;

    { if export succeeded and report can create its own dialogs, then inform the user of success }
    if (success and FConfirmDialog) then
      ShowConfirmDialog(self, filename, etFixed);
  end;
end;


{--------------------------}
{ EXPORT TO DELIMITED TEXT }
{--------------------------}
procedure TFlexiReport.ExportReportWithDelimiter(filename :string; delim :string);
var
  iGrid,
  Row,
  Col :integer;
  outfile :TFileStream;
  CurrentGrid :TStringGrid;
  charno :integer;
  success :boolean;
const
  CarriageReturn :string[2] = #13#10;
begin
  success:=True;

  try
    outfile := TFileStream.create(filename, fmCreate);
    try
      for iGrid := 0 to FGrids.count-1 do
      begin
        CurrentGrid := TStringGrid(FGrids.Items[iGrid]);
        for Row := 0 to CurrentGrid.RowCount - 1 do
        begin
          for Col := 0 to CurrentGrid.ColCount - 1 do
          begin
            { populate the ExportInfo structure }
            with FExportInfo do
            begin
              Grid := CurrentGrid;
              CellText :=  CurrentGrid.Cells[Col, Row];
              iRow := Row;
              iCol := Col;
              TextJustify := DT_LEFT;
              CellType := ctString;
              ExportType := etDelimited;
            end;

            { trigger the OnExport event handler if required }
            if assigned(FOnExportCell) then
            begin
              { trigger the event, returning the revised celltext and textjustify }
              FOnExportCell(self, FExportInfo);
            end;

            { write the string out to the stream, excluding the byte for strings size }
            for charno := 1 to Length(FExportInfo.CellText) do
              outfile.write(FExportInfo.CellText[charno], 1);

            outfile.write(delim[1], 1);
          end;
          outfile.write(CarriageReturn[1], 1);
          outfile.write(CarriageReturn[2], 1);
        end;
      end;
    finally
      outfile.Free;
//      GP_RestoreCursor(Self);
      if (success) then
        ShowConfirmDialog(self, filename, etDelimited);
    end;
    except
      on EFCreateError do
      begin
        MessageDLG('Error: There was an error trying to save to the specified file.' + #13#13 +
                   'Solution: Ensure that the file is not in use, is writable, and' +
                   ' the filename is valid.', mtError, [mbOK], 0);
      end;
  end;
end;

procedure TFlexiReport.SetSaveDialogFilters(aSaveDialog :TSaveDialog; ExportType :PRN_TExportType);
begin
  with aSaveDialog do
  begin
    { retrieve what type of text file the user wishes to produce }
    case ExportType of
      etExcel:
      begin
        Filter := 'Excel WorkSheet (*.XLS)|*.xls|';
        DefaultExt := 'xls';
        Filename := '*.xls';
      end;

      etFixed:
      begin
        Filter:='Text - Fixed Column (*.TXT)|*.txt|';
        DefaultExt := 'txt';
        Filename := '*.txt';
      end;

      etDelimited:
      begin
        { if a comma delimited file has been chosen, enable CSV in the filters }
        if (copy(ExportOptionsDialog.DelimiterEditBox.Text, 1, 1)=',') then
        begin
          Filter := 'Text (*.TXT)|*.txt|' +
                    'Comma Delimited Text (*.CSV)|*.csv|';
          DefaultExt := 'csv';
          Filename := '*.csv';
        end
        else
        { if not CSV, then add TXT extension }
        begin
          Filter := 'Text (*.TXT)|*.txt';
          DefaultExt := 'txt';
          Filename := '*.txt';
        end;
      end;
    end; { end case ExportType }

    { in all cases we should let 'All Files' be in the filter }
    Filter := Filter + 'All files (*.*)|*.*|';
    FilterIndex := 1;
  end;
end;

procedure TFlexiReport.ShowConfirmDialog(AOwner :TComponent; Filename :string; ExportType :PRN_TExportType);
var
  ExportConfirmDlg :TExportConfirmDlg;
begin
  ExportConfirmDlg := nil;
  try
    ExportConfirmDlg := TExportConfirmDlg.create(self, Filename, ExportType);
    ExportConfirmDlg.ShowModal;
  finally
    if (ExportConfirmDlg <> nil) then
      ExportConfirmDlg.Free;
  end;  
end;

end.
