{--------------------------------------------------------------------------------}
{                                                                                }
{                            UNIT : PRINTREP                                     }
{                                                                                }
{   TITLE       : Print_Report                                                   }
{                                                                                }
{   EXPLANATION : A class which encapsulates the printing of a report, from the  }
{                 report viewer.                                                 }
{                                                                                }
{   DATE        : 27/10/96                                                       }
{   AUTHOR      : P.CHAPMAN                                                      }
{   STATUS      : Development                                                    }
{                                                                                }
{--------------------------------------------------------------------------------}
{  Modifications                                                                 }
{                                                                                }
{  10/01/97 PAC - Implemented Fit To Page option, preventing horizontal pagespreads }
{                                                                                }
{  01/05/97 PAC - Introduced Printing structures and events allowing reports to  }
{                   to be printed using the FlexiReport class, ie without the    }
{                   need for a report viewer                                     }
{                                                                                }
{ WBF136.PAC - Added HeaderFontIndex implementation. When specified, in          }
{              the PrintInfoStructure, prints the header and description using   }
{              the font of the canvas, of the stringgrid at that index.          }
{                                                                                }
{================================================================================}
(*
  EXPLANATION:
  The main printing loop is as follows:

  if a report is not more than a page wide then grids are printed in order and
  pagenumbers increase as each vertical page is completed.

  If a report is more than one page in width, then each grid is printed in order,
  but only the section of the grid that fits horizontally onto the page width.
  When all grids are done, the loop is repeated, printing each grids section that
  fits on the second page. So for a report which spans two page spreads horizontally,
  the order of pages is this:

    -- --  Page numbers increase by 1 for the vertical
    |1|4|  and 0.1 for the horizontal, thus top left page = 1.1
    -- --  and bottom right page in the example is 3.2.
    |2|5|
    -- --
    |3|6|
    -- --

  The amount of each grid that will fit on each page is precalculated and stored in
  FPageSpreads, which is a collection of collections of Column records for each column
  for each grid. Each column record stores the corresponding columns maximum width,
  and which pagespread this column will appear on. The width of each column is
  multiplied by a scaling factor before bing stored in the corresponding
  ReportColumn record. This is not to account for screen to printer resolution scaling,
  but is used for the fit to page option. Scaling Factor is initially 1.00.

  If the report spans more than one page horizontally, and the fit to page option is ON,
  then the pagespread is recalculated until the highest pagespread of any grid is 1.
  Each time the pagespreads are calculated, if the highest pagespread > 1 then the
  scaling factor is reduced by the FScalingGranularity value, and the pagespreads are
  recalced. The scaling granularity is initially set to 0.001 which very accurately
  squashes the grids until they fit on the page exactly. This may prove to be slow
  for huge fit to page grids, Hence it can be changed. the smaller the number, the slower
  the calculations will take to fit to page, but the more optimum the fit will be.

  ScalingGranularity always lies between 0.5 and 0.0001. A granularity of 0.5 would
  keep halving the width of a large grid until it fit onto the page, but the grid
  would likely be squashed smaller than the optimum size.

  There are two structures which are used to control the output of the printing
  and are intended to be modified on the fly, with the use of the Printing events
  which have been defined...

  OnPrintCell
  OnPrintNewGrid
  OnPrintNewPage
  OnPrintGetColWidth
*)

unit Printrep;

interface

uses
  Classes, Printers, Grids,
  WinProcs, WinTypes, Dialogs,
  SysUtils, Graphics, PRN_Type;

type
 TReportColumn = class(TObject)
  private
    FMaxWidth :longint;
    FPageSpread :Integer;
  public
    property PageSpread :Integer   read FPageSpread   write FPageSpread;
    property MaxWidth :longint     read FMaxWidth     write FMaxWidth;
    constructor create;
  end;

  TPrintedReport = class(TObject)
  private
    FShadeRect :TRect;

    FGrids :TList; { collection of TStringGrids which the report is held in }
    FiGrid :integer;

    FPageSpreads :TList; { each item is a collection of ReportColumn objects for each grid }
    FXpos, FYpos :longint; { co-ordinates of the cursor on the output canvas }

    FDrawInfo,                              { structure to define how the next cell is to be drawn }
    FDefaultDrawInfo :PRN_TDrawInfoStruct;  { used to hold defaul DrawInfo }

    FCurrentGrid :TStringGrid; { the currently report string grid, held within FGrids }
    FGridWidth :longint;       { the grid dimensions, scaled to activecanvas pixels }
    FPixPerInX,
    FPixPerInY :real; { pixels per logical inch for the active canvas }

    FPageNoPosn :Integer; { justification of the page no at footer of page }
    FAborted :Boolean; { set when the print job is cancelled }

    FCurrentSpread :Integer; { which horizontal page spread are we processing }
    FMaximumSpread :integer; { the pagespread of the widest grid }
    FWidestGridIndex :integer; { index into the collection of grids, to the widest grid }
    FWidestGridWidth :longint; { width of the widest grid in screen pixels }

    FStartCol, FEndCol :Integer; { range of cols on current page spread }
    FPageNumber :longint; { major page number, ie vertical numbering }
    FScaling_Factor :real;  { set the scaline factor (1.00 if fit to page is off }

  { * only available after first page is printed }
    FUsablePageHeight :longint; { page height available for printing non-fixed rows }
    FTotalNumberOfPages :integer;  { total number of pages which the report will take up }
    FStartOfUsableHeight :longint; { Ypos after fixed header is printed }
    FEndOfUsableHeight :longint;   { YPos after last row is printed }

    FDateTime :TDateTime; { holds the date and time (so time doesnt change across pages) }
    FPrintInfo :PRN_TPrintInfoStruct;

    { holds the number of pages each grid will span }
    { the highest value in this list is recorded in FMaximumSpread }
    FMaxSpread :TStringList;

    FPrinting :Boolean;
    FScalingGranularity :real;

    FFooterBitmap :TBitmap;
    FCanvas :TCanvas;

    { holds the current font while the headerfont is being used } 
    FOldFont :TFont;

    { WYSIWYG oriented routines }
    procedure GetPrinterResolutions;                   { retrieve vert/horz aspects and resolutions }
    procedure SetPrinterFont(aScreenFont :TFont);      { scales screen font to printer font }

    function getSpaceLeft :longint;  { returns amount of pixels left to print in }
    procedure PrintDescription;      { print report description }
    procedure PrintFixedRows;        { print column headers }
    procedure PrintHeaderText;       { print report header }
    function CheckNewPage(proposedFYpos :longint; PrintFixed: Boolean) :Boolean; { start new page if needed }
    procedure setPageNoPosn(pos :Integer);
    function EstimateTotalPages :integer;
    procedure GenericDrawCell(FDrawInfo :PRN_TDrawInfoStruct);
    procedure BuildColumnWidthsList;
    procedure setDefaultDrawInfo(DrawInfo :PRN_TDrawInfoStruct);
    procedure SetColumnScope;
    procedure DrawGridLines(aRect :TRect);

    procedure InstallHeaderFont;
    procedure UnInstallHeaderFont;

    procedure ValidateHeaderFontIndex(var PrintInfo :PRN_TPrintInfoStruct);

   { property to calculate the amount of vertical space left to print on page }
    property SpaceLeft :longint read getSpaceLeft;

  public

    constructor Create(aOwner :TObject; GridList :TList);
    destructor Destroy; override;

    { prints the current report }
    procedure Print;

    { stop processing and cancel print job }
    procedure Abort;

    procedure EndPrintPage(StartNew :boolean); { print footer, and tidy up page }
    function ScaleXToCanvas(xValue :longint) :longint; { converts screenX to printerX }
    function ScaleYToCanvas(yValue :longint) :longint; { converts screenY to printerY }

    { printer output options }
    property PageNoPosn :Integer                  read FPageNoPosn         write setPageNoPosn;
    property Aborted :Boolean                     read FAborted            write FAborted;
    property PrintInfo :PRN_TPrintInfoStruct      read FPrintInfo          write FPrintInfo;
    property DefaultDrawInfo :PRN_TDrawInfoStruct read FDefaultDrawInfo    write setDefaultDrawInfo;
    property Printing :Boolean                    read FPrinting;          { R/O }
    property ScalingGranularity :real             read FScalingGranularity write FScalingGranularity;

    property Canvas :TCanvas                      read FCanvas;
  end;

implementation

const
  REP_HIGHTEXT = 'jyMgI';    { Tallest string, used to calc height needed for a row }

procedure TPrintedReport.setPageNoPosn(pos :Integer);
{ property write method for PageNoPosn }
begin
  if ((pos = DT_CENTER) or
      (pos = DT_LEFT) or
      (pos = DT_RIGHT)) then
  FPageNoPosn := pos;
end;

procedure TPrintedReport.Abort;
begin
{  Printer.Abort; }
  Aborted := True;
end;

constructor TReportColumn.create;
{ TReportColumn, used to hold the max width reqd for a column }
begin
  inherited create;
  MaxWidth := 0;
  PageSpread := 0;
end;

constructor TPrintedReport.Create(aOwner :TObject; GridList :TList);
{ Creates an instance of a printed report, taking a list of string grids as a parameter }
begin
  inherited create;

  FGrids := GridList;
  FiGrid := 0;
  FPageNoPosn := DT_CENTER;
  FAborted := False;
  FScaling_Factor := 1.00;
  FScalingGranularity := 0.001;
  FPageNumber := 1;
  FPrinting := False;
  FWidestGridIndex := 0;
  FWidestGridWidth := 0;

 { record the date and time the report was started, so that time does not change }
 { as the report prints each page }
  FDateTime := Now;

 { list to hold the pagespreads for each of the grids }
  FMaxSpread := TStringList.create;

 { assume that the widest grid spans only 1 page, until figured otherwise }
  FMaximumSpread := 1;

 { take PrintInfo from the parent FlexiReport }
  FPrintInfo := PrintInfo;

  { create a font to use for swapping }
  FOldFont := TFont.create;

  FFooterBitmap := nil;
end;

destructor TPrintedReport.Destroy;
begin
  { destroy the list of maximum page spreads for each grid }
  FMaxSpread.Free;

  { destroy the bitmap if present }
  if assigned(FFooterBitmap) then
    FFooterBitmap.Free;

  if assigned(FOldFont) then
    FOldFont.Free;

  inherited destroy;
end;

function TPrintedReport.CheckNewPage(proposedFYpos :longint; PrintFixed :Boolean) :Boolean;
{ creates new page if proposed YPos exceeds page height. returns True on newpage }
begin
  CheckNewPage := False;
  if ((Printer.PageHeight - ProposedFYpos) <
      (2 * Printer.Canvas.TextHeight('Page xx'))) then
  begin
    CheckNewPage := True;

    { if we are just ending the first page then set some private values }
    if (FPageNumber = 1) then
    begin
      FEndOfUsableHeight := FYPos;
      FUsablePageHeight := FEndOfUsableHeight - FStartOfUsableHeight;
      FTotalNumberOfPages := EstimateTotalPages;
    end;

    { finish the page by printing the footer }
    EndPrintPage(True);
    inc(FPageNumber);

   { if FPrintInfo.ShowHeader then }
      PrintHeaderText;

    { if there are more rows to print on the current grid }
    { then print the fixed rows for the current grid }
    { otherwise leave the grid to print its own FixedRows }
    if (PrintFixed and (not FPrintInfo.OmitFixedRows)) then
      PrintFixedRows;
  end;
end;

function TPrintedReport.getSpaceLeft :longint;
{ read access method for property }
begin
  getSpaceLeft := Printer.PageHeight - FYpos -
                 (2 * Printer.Canvas.TextHeight('Page 99'));
end;

procedure TPrintedReport.PrintHeaderText;
{ Prints the report header text, with horizontal line, increases FYpos }
{ used to start each fresh page }
var
  Bound :TRect;
  TextOption :longint;
begin
  TextOption := DT_NOPREFIX or DT_SINGLELINE or DT_NOCLIP or DT_VCENTER;

  { if a header has been assigned }
  if ((FPrintInfo.ShowHeader) and (FPrintInfo.Header <> '')) then
  begin
    { install header font }
    InstallHeaderFont;

    { if header text is wider than the widest grid, then header across top of page }
    if (FDrawInfo.PrintStatus.ActiveCanvas.TextWidth(FPrintInfo.Header) >
        ScaleXToCanvas(Printer.PageWidth)) then
    begin
      TextOption := TextOption or DT_LEFT;
      Bound := Rect(ScaleXToCanvas(5), FYpos, Printer.PageWidth, FYpos +
                      Trunc(1.5 * FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(FPrintInfo.Header)));
    end
    else
    begin
      TextOption := TextOption or DT_CENTER;
      Bound := Rect(0, FYpos, Printer.PageWidth, FYpos +
                      Trunc(1.5 * FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(FPrintInfo.Header)));
    end;

   { draw a temporary bounding box }
    FDrawInfo.PrintStatus.ActiveCanvas.Rectangle(Bound.Left, Bound.Top, Bound.Right, Bound.Bottom);

   { Actually draw the text into the bounding area }
    DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, @FPrintInfo.Header[1],
      Length(FPrintInfo.Header), Bound, TextOption);

   { Increment FYpos by the height used up by the report header }
    FYPos := FYPos + (Bound.Bottom - Bound.Top) +
               Trunc(FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(REP_HIGHTEXT) / 2);

    { reinstall the font which was being used before the header was printed }
    UnInstallHeaderFont;
  end;

  { if an event handler has been assigned to the OnNewPage event, and we can estimate }
  { the total pages (ie pageNo > 1),  then trigger the event }
  if (assigned(FPrintInfo.OnPrintNewPage) and (FPageNumber > 1)) then
    FPrintInfo.OnPrintNewPage(self, FPageNumber, FTotalNumberOfPages, FXPos, FYPos);
end;

procedure TPrintedReport.EndPrintPage(StartNew :boolean);
{ prints the footer text and page number. Starts a new page is requested }
var
  FooterStr :string;
  DateTimeStr :string;
  Bound, origBound :TRect;
  aspect :real;
  bitmapleft :longint;
begin
 { Load footer into a Pascal string and transfer to a PChar }
  DateTimeStr := FormatDateTime(' ddd dd/mm/yy hh:mm ', FDateTime);

  { create the text to go on the bottom of the page. This includes page number in }
  { the form of page v.h, where v is the vertical page number and h is the }
  { horizontal page number (otherwise known as the pagespread). The date and time }
  { is also appended to the footer }
  FooterStr := 'Page ' + IntToStr(FPageNumber);

  if (FMaximumSpread > 1) then
    FooterStr := FooterStr + '.' + IntToStr(FCurrentSpread);

  FooterStr := FooterStr + ' - ' + DateTimeStr;

  { Set the current font, to the font of the canvas of the grid at the position }
  { specified in PrintInfo.HeaderFontIndex }
  InstallHeaderFont;

 { if footerwidth > report width then just indent, as centering this would lose some of the text }
 { just indent the footer, and draw it left justified }
  if ((FDrawInfo.PrintStatus.ActiveCanvas.TextWidth(footerstr) > FGridWidth) and
      (PageNoPosn = DT_CENTER)) then
    Bound := Rect(ScaleXToCanvas(5), Printer.PageHeight - (2*FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(FooterStr)),
                   Printer.PageWidth, Printer.PageHeight)
  else
   { Pass boundaries as grid width, and center if selected in PageNoPosn }
    Bound := Rect(0, Printer.PageHeight - (2*FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(FooterStr)),
                   Printer.PageWidth, Printer.PageHeight);

   { take a copy of the bounding box for the footer bitmap }
    origBound := Bound;

   { Actually draw the footer text }
    DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, @FooterStr[1], Length(FooterStr), Bound,
               PageNoPosn or DT_NOPREFIX or DT_SINGLELINE or DT_BOTTOM);

  if assigned(FFooterBitmap) then
  begin
    { get the ratio of height to width of the footer bitmap }
    aspect := FFooterBitmap.Height / FFooterBitmap.Width;

    { we know the actual height of the bounding box }
    { so we calculate the width required by the bitmap if it is scaled to this height }
    bitmapleft := origBound.Right - Round((origBound.Bottom - origBound.Top) / aspect);

    { modify the bounding box }
    origBound.Left := bitmapleft;

    { then draw it right justified in the footer }
    FDrawInfo.PrintStatus.ActiveCanvas.StretchDraw(origBound, FFooterBitmap);
  end;

  { reinstall the font which was being used before the header was printed }
  UnInstallHeaderFont;

  { start a new page }
  if (StartNew) then
  begin
    Printer.NewPage;

   { set the cursor to the top left of the new page }
    FXpos := 0;
    FYpos := 0;
  end;
end;

procedure TPrintedReport.PrintFixedRows;
{ Prints the actual fixed header rows containing the column labels }
{ used at the start of each fresh page, just after the headertext }
{ comes after the description text on the very first page }
var
  irow, icol :Integer;
  tmpXpos :longint;
  StretchedHeight :longint;
  ThisRowHeight :longint;
  ThisColWidth :longint;
  PrintingBreakCols :Boolean;
begin
  for irow := 0 to FCurrentGrid.FixedRows-1 do
  begin
   { Start from the left of the page }
    tmpXpos := 0;

   { Calc the current row height, and scale to equivalent size on printers canvas }
    ThisRowHeight := Round(ScaleYToCanvas(FCurrentGrid.RowHeights[irow]) * FScaling_Factor);
    StretchedHeight := ThisRowHeight;

   { start off printing the columns for the fixed rows (breakdown columns) }
    icol := 0;
    PrintingBreakCols := True;

    { for each column that needs to be included on this page }
    while (icol <= FEndCol) do
    begin
      { if all breakdown cols are printed, start printing appropriate non-breakdown cols }
      if ((icol > FCurrentGrid.FixedCols - 1) and
          (PrintingBreakCols)) then
      begin
        iCol := FStartCol;
        PrintingBreakCols := False;
      end;

      ThisColWidth := ScaleXToCanvas(TReportColumn(TList(FPageSpreads.Items[FiGrid]).Items[icol]).MaxWidth);

      { fill in the drawing attributes structure with the defaults supplied }
      FDrawInfo := FDefaultDrawInfo;

      { now set some calculated fields }
      with FDrawInfo do
      begin
        PrintStatus.Row := irow;
        PrintStatus.Col := icol;
        PrintStatus.GridIndex := FiGrid;
        PrintStatus.CurrentGrid := TStringGrid(FGrids.Items[FiGrid]);
        PrintStatus.ActiveCanvas := Printer.Canvas;
        CellText := PrintStatus.CurrentGrid.Cells[PrintStatus.Col, PrintStatus.Row];
      end;

      FDrawInfo.CellRect := Rect(tmpXpos, FYpos, tmpXpos + ThisColWidth, FYpos + ThisRowHeight);

      { take a copy of the cell rect so that if the cell is shaded, we shade }
      { the whole cell, and not the modified rect returned from the OnPrintCell handler }
      FShadeRect := FDrawInfo.CellRect;

      { shift the X position along by the width of the column }
      inc(tmpXpos, ThisColWidth);

      { trigger the event, allowing the user to modify the drawing info }
      { before the current cell is printed }
      if (assigned(FPrintInfo.OnPrintCell)) then
        FPrintInfo.OnPrintCell(self, FDrawInfo);

      { if gridlines are on then draw them }
      if (FPrintInfo.ShowGridLines) then
        DrawGridLines(FShadeRect);

      { call the drawcell using the newly modified parameters, unless disabled }
      if (FDrawInfo.DefaultDrawing) then
        GenericDrawCell(FDrawInfo);

      { start new page if required }
      if (FPrintInfo.StartNewPage) then
      begin
        CheckNewPage(Printer.PageHeight + 1, False); 
        FPrintInfo.StartNewPage := False;
      end;

      { move onto the next column }
      inc(iCol);
    end;

    { move down the page by the amount which was actually used by this row }
    FYpos := FYpos + StretchedHeight;
  end;
  FStartOfUsableHeight := FYPos;
end;

procedure TPrintedReport.PrintDescription;
var
  DescSize,                    { size of currently remaining description }
  DescLoop,
  CharCount,
  DescPrinted,                 { no of chars of description printed }
  DescTest,                    { no of chars tested to see if can be printed }
  CharNum :longint;            { loop used to copy strings to buffer }
  Text,
  StartingText :PChar;
  TempString :string;
  Bound :TRect;
begin
  if (FPrintInfo.Description <> nil) then
  begin
    if ((FPrintInfo.ShowDescription) and (FPrintInfo.Description.Count > 0)) then
    begin
      InstallHeaderFont;

     { calculate the length of the report description }
      DescSize := 0;
      for DescLoop := 0 to FPrintInfo.Description.Count - 1 do
        DescSize := DescSize + Length(FPrintInfo.Description.Strings[DescLoop]) + 1;

     { create a PChar big enough to hold description text }
      Text := StrAlloc(DescSize + 1);
      FillChar(Text^, DescSize + 1, #0);

      { record where the buffer starts, to dispose of later (Text will change) }
      StartingText := Text;

     { copy each character from the stringlist into the PChar }
      CharNum := 0;

      for DescLoop := 0 to FPrintInfo.Description.Count - 1 do
      begin
        TempString := FPrintInfo.Description.Strings[DescLoop];
        for CharCount := 1 to Length(TempString) do
        begin
          Text[CharNum] := TempString[CharCount];
          inc(CharNum);
        end;
        Text[CharNum] := ' ';
        inc(CharNum);
      end;

      Text[charnum] := #0;

      DescPrinted := 0;

     { keep going till all of description has been printed }
      while (DescPrinted < DescSize) do
      begin
       { create a bounding box the width of the page and one pixel high }
        Bound := Rect(0, FYpos, Printer.PageWidth, FYpos + 1);

       { stretch the bounding box if necessary }
        DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, Text, -1, Bound,
          DT_LEFT or DT_NOPREFIX or DT_WORDBREAK or DT_CALCRECT or DT_EXPANDTABS);

        DescTest := DescSize - DescPrinted;

       { while bounding box does not fit on remaining page space }
        while ((Bound.Bottom - Bound.Top) > SpaceLeft) and (DescTest > 0) do
        begin
         { knock a word off the PChar }
          while ((DescTest > 0) and (Text[DescTest] <> ' ')) do
            Dec(DescTest);

         { skip the space we have landed on }
          if (Text[DescTest] = ' ') then
            Dec(DescTest);

         { see if the description can now be wrapped into current page }
          DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, Text, DescTest, Bound,
            DT_LEFT or DT_NOPREFIX or DT_WORDBREAK or DT_CALCRECT or DT_EXPANDTABS);

        end;

        { At this point the first DescTest chars of the PChar will fit on the }
        { current page. Actually draw the text }
        if (DescTest > 0) then
        begin
          { if bounding box would not fit on page, start new page, and modify }
          { the bounding rect to account for check if we need a new page for this bounding box, with no fixed headers }
          {if (not }
            CheckNewPage(FYpos + (Bound.Bottom - Bound.Top), False);

          {  then FYpos := FYpos + (Bound.Bottom - Bound.Top); }
          (*else
          begin
          { modify the bounding rectangle to start at top of the new page }
            Bound.Bottom := FYPos + (Bound.Bottom - Bound.Top);
            Bound.Top := FYpos;
          end; *)

          { only actually draw the text we are on the first page spread }
          { otherwise all this calculating code is called anyway, and a space }
          { is left where the description would have appeared }
          if (FCurrentSpread = 1) then
            DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, Text, DescTest, Bound,
              DT_LEFT or DT_NOPREFIX or DT_WORDBREAK or DT_EXPANDTABS);

          DescPrinted := DescPrinted + DescTest;

         { if there is still some description text to be printed }
          if (DescPrinted < DescSize) then
          begin
            { move the pointer into the PChar by the amount printed }
            Text := Text + DescPrinted;

            { update the new 'size' of the string (not of the PChar though) }
            DescSize := DescSize - DescPrinted;

            { pretend we havent printed any of the description yet }
            DescPrinted := 0;
          end;
        end;

        (* else
        begin
          DescPrinted := DescSize;
          { start new page here, so that the rest of the description may be printed }
        end; *)

        { if all of the description fit on the page OK then update the Y-cursor position }
        if (not CheckNewPage(FYpos + (Bound.Bottom - Bound.Top), False)) then
           FYpos := FYpos + (Bound.Bottom - Bound.Top);

        { force a new page if there is no space for another line of text }
  {      if (not CheckNewPage(FYpos +  Trunc(FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(REP_HIGHTEXT)), False); }
        (*{else }
        begin
          { the call to checknewpage started a new page, so move the bounding box to the }
          { top of the sheet }
        { modify the bounding rectangle to start at top of the new page }
          Bound.Bottom := FYPos + (Bound.Bottom - Bound.Top);
          Bound.Top := FYpos;
        end;
        { start a new page for the remaining description if needed }
        if not (CheckNewPage(FYpos + Trunc(FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(REP_HIGHTEXT)/2), False)) then
        begin
          FYpos := FYpos + Trunc(FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(REP_HIGHTEXT)/2);
        end; *)
      end;

     { free up the main PChar after it has been swapped }
      StrDispose(StartingText);
{      Text := nil; }
{      StartingText := nil; }
      UnInstallHeaderFont;
    end;
  end;
end;


function TPrintedReport.ScaleXToCanvas(xValue :longint) :longint;
{ given a horizontal pixel position for the screen, scales it to the canvas }
begin
  If FPrinting then
    result := Trunc((xValue / FPrintInfo.PixelsPerInch) * FPixPerInX)
  else
    result := xValue;
end;

function TPrintedReport.ScaleYToCanvas(yValue :longint) :longint;
{ given a vertical pixel position for the screen, scales it to the canvas }
begin
  if FPrinting then
    ScaleYToCanvas := Trunc((yValue / FPrintInfo.PixelsPerInch) * FPixPerInY)
  else
    result := yValue;
end;

procedure TPrintedReport.Print;
{ Prints the report to the printer. }
var
  StretchedHeight :longint;
 { holds the height of a string after it has been wordwrapped }
  gridno :longint;
 { index to the currently processing grid }
  irow, icol :longInt;
 { index for the current grids row and column }
  ThisColWidth :longint;
  ThisRowHeight :longint;
 { Current columns width, and rows height }
  DrawLoop, TmpLoop :Integer;

  LongestCell :integer;
  LongestCellLen :longint;
  CellLoop :integer;
  HeightReqd :longint;
  TempCellStr :string;

  SpreadLoop :Integer;
  TempColWidth :longint;
  TempPageSpread :Integer;

  PrintingBreakCols :Boolean;
  TempLong :longint;
  Bound :TRect;
  CellRect :TRect;
  OldBrushColor :TColor;
  OldBrushStyle :TBrushStyle;

  tmpRect :TRect;
  tmpCollect :TList;
  tmpRepCol :TReportColumn;

begin
  FDrawInfo.PrintStatus.ActiveCanvas := Printer.Canvas;

 { create list to hold list of maxwidths for each grid }
  FPageSpreads := TList.create;

 { set pointer to first grid }
  FCurrentGrid := TStringGrid(FGrids.Items[0]);

 { Build list of max column widths for each grid }
  BuildColumnWidthsList;

 { set the title of the print job }
  Printer.Title := PrintInfo.Header;

 { start a new print job }
  Printer.BeginDoc;

  FPrinting := True;
  FCanvas := Printer.Canvas;

 { get printer resolutions }
  GetPrinterResolutions;

 { find the widest grid in the collection }
 { then if fit to page is selected, use this grid to set the scaling factor }
 for GridNo := 0 to FGrids.Count - 1 do
 begin
   TempColWidth := 0;     { total column width for current grid }

   for icol := 0 to TStringGrid(FGrids.Items[GridNo]).ColCount - 1 do
     TempColWidth := TempColWidth + TStringGrid(FGrids.Items[GridNo]).ColWidths[icol];

   if (TempColWidth > FWidestGridWidth) then
   begin
     FWidestGridIndex := GridNo;
     FWidestGridWidth := TempColWidth;
   end;
 end;

 { Decide which page each grids column will appear on, setting the pagespread }
 { if FitToPage is selected, then scale all grids so that the widest grid will }
 { fit on the page. }
  for Gridno := 0 to FGrids.Count - 1 do
  begin
    repeat
     { calculate which cols go on which page, and set PageSpread property. }
     { Fixed columns have a page spread of zero, and hence get printed on every spread page }
      TempColWidth := 0;
      TempPageSpread := 1;

     { reset the maximum page spread, incase the page is being scaled down }
      FMaximumSpread := 1;

      for iCol := 0 to TStringGrid(FGrids.Items[GridNo]).ColCount-1 do
      begin
        { add the current column width onto amount of page width used up }
        TempLong := ScaleXToCanvas(Round(Fscaling_Factor *
          TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).MaxWidth));

        TempColWidth := TempColWidth + TempLong;
        if (TempColWidth >= (Printer.PageWidth - 20)) then
        begin
         { move onto next page spread }
          inc(TempPageSpread);

         { set the width on the new page taken up by the current column }
          TempColWidth := TempLong;

         { now add the col widths of the fixed columns as they appear on evey page }
          for TmpLoop := 0 to TStringGrid(FGrids.Items[GridNo]).FixedCols - 1 do
          begin
            TempLong := ScaleXToCanvas(TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[TmpLoop]).MaxWidth);
            TempColWidth := TempColWidth + TempLong;
          end;
        end;

        if (iCol < TStringGrid(FGrids.Items[GridNo]).FixedCols) then
          TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).PageSpread := 0
        else
          TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).PageSpread := TempPageSpread;

        { set the number of pages the report spreads across (horizontal pagespread) }
        FMaxSpread.Strings[GridNo] := IntToStr(TempPageSpread);

        { record the highest pagespread in the report }
        if (TempPageSpread > FMaximumSpread) then
          FMaximumSpread := TempPageSpread;
      end;

      { decrease the scaling factor until the report fits to the page }
      if (FPrintInfo.FitToPage and (FMaximumSpread > 1)) then
        FScaling_Factor := FScaling_Factor - FScalingGranularity;

    until ((TempPageSpread <= 1) or (not FPrintInfo.FitToPage));

    { if the scaling_factor is not 1.00 then it means that the columns and text }
    { has been shrunk to fit onto a page, so we should update the column widths }
    if (FPrintInfo.FitToPage and (Fscaling_factor < 1.00)) then
    begin
      for icol := 0 to TList(FPageSpreads.Items[GridNo]).count - 1 do
      begin
        TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).MaxWidth  :=
          Round(TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).MaxWidth * Fscaling_factor);
      end;
    end;
  end;

 { Add up the maximum width of the grid }
  for GridNo := 0 to FGrids.count - 1 do
  begin
    FGridWidth := 0;
    for icol := 0 to TList(FPageSpreads.Items[GridNo]).Count - 1 do
      FGridWidth := FGridWidth + TReportColumn(TList(FPageSpreads.Items[GridNo]).Items[icol]).MaxWidth;
  end;

  FXpos := 0;
  FYpos := 0;

 { Use first grid in report to retrieve header }
  FDrawInfo.PrintStatus.GridIndex := 0;

{ ------------------------------------------- }
{ THE FIRST OUTPUT TO THE PRINTER OCCURS HERE }
{ ------------------------------------------- }
  FCurrentSpread := 1;

  { Print the report header }
  PrintHeaderText;

 { ### MAIN LOOP ### }
  repeat
    FPageNumber := 1;

    For GridNo := 0 to FGrids.count - 1 do
    begin
      { set an index and a pointer to the current grid }
      FiGrid := GridNo;
      FCurrentGrid := TStringGrid(FGrids.Items[GridNo]);

      { use the font taken from the current grid }
      SetPrinterFont(TStringGrid(FGrids.Items[GridNo]).Canvas.Font);

      { calculate the range of columns to be printed. }
      { this sets FStartCol and FEndCol. Fixed columns are printed regardless. }
      SetColumnScope;

      { trigger the OnNewGrid event, allowing the user to manipulate the parameters before }
      { any of this new grid is printed. This may include a change of description }
      { as the description is printed for each grid }
      if assigned(FPrintInfo.OnPrintNewGrid) then
        FPrintInfo.OnPrintNewGrid(Self, FiGrid);

      { start new page if required }
      if (FPrintInfo.StartNewPage) then
      begin
{        EndPrintPage(True); }
        CheckNewPage(Printer.PageHeight + 1, False); { EndPrintPage; }
        FPrintInfo.StartNewPage := False;
      end;

      { if on 2nd or further grid, and gridspacing is set }
      if ((FPrintInfo.GridSpacing > 0) or (FPrintInfo.GridSpacing = -1)) then
      begin
        if (FiGrid > 0) then
          { if we can fit the spacing on this page then move down the correct amount }
          if not CheckNewPage(FYPos + ScaleYToCanvas(FPrintInfo.GridSpacing), False) then
            FYpos := FYpos + ScaleYToCanvas(FPrintInfo.GridSpacing);
      end;

      { Print the description if required (ie (PageSpread = 1) & (description present)) }
      PrintDescription;

{      SetPrinterFont(FCurrentGrid.Font); }

      { set the readonly properties to reflect current cell to be printed }
      with FDrawInfo.PrintStatus do
      begin
        GridIndex := FiGrid;
        CurrentGrid := FCurrentGrid;
        ActiveCanvas := Printer.Canvas;
      end;

      if (not FPrintInfo.OmitFixedRows) then
        PrintFixedRows;

      FGridWidth := 0;
      for TmpLoop := 0 to FCurrentGrid.ColCount-1 do
        FGridWidth := FGridWidth + FCurrentGrid.ColWidths[TmpLoop];

      { for each row to be printed }
      iRow := FCurrentGrid.FixedRows;
      while (iRow <= FCurrentGrid.RowCount - 1) do
      begin
       { Reset the stretched height back to zero }
{        StretchedHeight := 0; }  {PAC}

       { Retrieve the longest cell text for that row }
        LongestCellLen := 0;
        LongestCell := 0;
        for CellLoop := 0 to FCurrentGrid.colcount-1 do
        begin
          if Length(FCurrentGrid.Cells[CellLoop, iRow]) > LongestCellLen then
          begin
            LongestCellLen := Length(FCurrentGrid.Cells[CellLoop, iRow]);
            LongestCell := CellLoop;
          end;
        end;

       { Calculate the height needed to display this text, using its column width }
        TempCellStr := FCurrentGrid.Cells[LongestCell, iRow];

       { if cell is blank then put a characters in just to act as height specifier }
        if (TempCellStr = '') then
          TempCellStr := REP_HIGHTEXT;

       tmpCollect := TList(FPageSpreads.Items[FiGrid]);
       tmpRepCol := TReportColumn(tmpCollect.Items[LongestCell]);

       { calculate the height required to display the cell }
        Bound := Rect(0, 0,
                  Round(FScaling_Factor * ScaleXToCanvas(tmpRepCol.MaxWidth)),
                  Round(FScaling_Factor * FDrawInfo.PrintStatus.ActiveCanvas.TextHeight(REP_HIGHTEXT)));

       { Calculate how much height is required to display text when wordwrapped }
        DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle,
                  @TempCellStr[1], Length(TempCellStr), Bound,
                  DT_LEFT or DT_NOPREFIX or DT_WORDBREAK or DT_CALCRECT);

       { Calc the standard row height, and scale to equivalent size on printers canvas }
        ThisRowHeight := Round(ScaleYToCanvas(FCurrentGrid.RowHeights[irow]) * FScaling_factor);

       { check if outputting this row would overflow the page. If so start a fresh }
       { page, and print the fixed rows along the top, before proceeding }
        CheckNewPage(FYpos + ThisRowHeight, True);

        FXpos := 0;

      { MAIN DRAWING LOOP: DRAWS EACH CELL FOR A ROW ON A PAGE }
        icol := 0;
        StretchedHeight := ThisRowHeight;
        PrintingBreakCols := True;

      { for each column up to the last column to be printed on this page }
        while (icol <= FEndCol) do
        begin
        { if finished printing the breakdown columns then start printing the normal columns }
          if ((iCol > TStringGrid(FGrids.Items[GridNo]).FixedCols - 1) and
              (PrintingBreakCols)) then
          begin
            PrintingBreakCols := False;
            if (iCol < FStartCol) then
              iCol := FStartCol;
          end;

          ThisColWidth := ScaleXToCanvas(TReportColumn(TList(FPageSpreads.Items[FiGrid]).Items[icol]).MaxWidth);

          { retrieve the cells text }
          TempCellStr := FCurrentGrid.Cells[iCol, iRow];

          { fill in the drawing attributes structure with the defaults supplied }
          FDrawInfo := FDefaultDrawInfo;

          { set the defaultDrawing Information }
          FDrawInfo.CellRect := Rect(FXpos, FYpos, FXpos + ThisColWidth, FYpos + ThisRowHeight);

          { take a copy of the original cell rect so that gridlines can be drawn later }
          CellRect := FDrawInfo.CellRect;

          { set the readonly fields in the PrintStatus subrecord of the DrawInfo structure }
          with FDrawInfo do
          begin
            PrintStatus.Row := irow;
            PrintStatus.Col := icol;
            PrintStatus.GridIndex := FiGrid;
            PrintStatus.CurrentGrid := TStringGrid(FGrids.Items[FiGrid]);
            PrintStatus.ActiveCanvas := Printer.Canvas;
            CellText := PrintStatus.CurrentGrid.Cells[PrintStatus.Col, PrintStatus.Row];
          end;

          inc(FXpos, ThisColWidth);

          { take a copy of the cell rect so that if the cell is shaded, we shade }
          { the whole cell, and not the rect returned from the OnPrintCell handler }
          FShadeRect := FDrawInfo.CellRect;

          { trigger OnPrintCell, allowing the user to modify the drawing info }
          if (assigned(FPrintInfo.OnPrintCell)) then
            FPrintInfo.OnPrintCell(self, FDrawInfo);

          { if gridlines are on then draw them FShowGridLines then }
          if (FPrintInfo.ShowGridLines) then
            DrawGridLines(FShadeRect);

          { if reqd, call the drawcell using the newly modified parameters }
          if (FDrawInfo.DefaultDrawing) then
            GenericDrawCell(FDrawInfo);

          { start new page if required }
          if (FPrintInfo.StartNewPage) then
          begin
            CheckNewPage(Printer.PageHeight + 1, False); 
            FPrintInfo.StartNewPage := False;
          end;

          if ((FDrawInfo.CellRect.Bottom - FDrawInfo.CellRect.Top) > StretchedHeight) then
            StretchedHeight := FDrawInfo.CellRect.Bottom - FDrawInfo.CellRect.Top;

          { move onto the next column }
            inc(iCol);
        end;

      { check if we need a new page }
        if (not CheckNewPage(FYpos + StretchedHeight, True)) then
          FYpos := FYpos + StretchedHeight;

        inc(iRow);
        if Aborted then break;
      end; { end - for each row in current grid }
      if Aborted then break;
      CheckNewPage(FYpos, True);
    end; { end - for each grid }
    if Aborted then break;

    { start new page, but only if there are more page spreads to do }
    if (FCurrentSpread < FMaximumSpread) then
      CheckNewPage(Printer.PageHeight + 1, False);

    { move onto the next pagespread }
    inc(FCurrentSpread);
  until (FCurrentSpread > FMaximumSpread);

  { set the title of the print job }
  { Printer.Title := PrintInfo.Header; }

  if (not Aborted) then
  begin
    EndPrintPage(False);
    Printer.EndDoc; { end the print job }
  end
  else
  begin
    Printer.Abort;
    Printer.EndDoc;
  end;

  FCanvas := nil;
  FPrinting := False;
end;

function TPrintedReport.EstimateTotalPages :integer;
var
{  totalrows :integer; }
  iGrid, iRow :integer;
  ThisGrid :TStringGrid;
  rowHeight, totalrowheight :longint;
  totalpages :integer;
begin
  totalRowHeight := 0;

  { scan through each row in each grid }
  for iGrid := 0 to FGrids.Count - 1 do
  begin
    ThisGrid := TStringGrid(FGrids.Items[iGrid]);

    { for each NON FIXED row in the current grid }
    for iRow := ThisGrid.FixedRows to ThisGrid.RowCount - 1 do
    begin
      rowHeight := ScaleYToCanvas(ThisGrid.rowHeights[iRow]);
      inc(totalRowHeight, rowHeight);
    end;
  end;

{  TotalRowHeight := ScaleYToCanvas(Round(TotalRowHeight * FScaling_Factor)); }

  { assume each grid will be spaced like this one }
  inc(TotalRowHeight, ScaleYToCanvas(FPrintInfo.GridSpacing * FGrids.Count));

  { now divide the total row height, by the height available for NON-FIXED rows on a page }
  if ((totalrowheight = 0) or (FUsablePageHeight = 0)) then
    totalPages := 0
  else
    totalPages := Round(totalrowheight / FUsablePageHeight);

  result := totalPages;
end;

procedure TPrintedReport.GenericDrawCell(FDrawInfo :PRN_TDrawInfoStruct);
var
  oldBrushStyle :TBrushStyle;
  oldBrushColor :TColor;
  drawoptions :longint;
  tmpRect :TRect;
begin
  { draw the current cell using information from the drawinfo structure }
  with FDrawInfo do
  begin
    { apply shading if required }
    if (coShaded in Options) then
    begin
      with PrintStatus.ActiveCanvas do
      begin
        OldBrushStyle := Brush.Style;
        OldBrushColor := Brush.Color;

        Brush.Style := bsSolid;
        Brush.Color := ShadeColor;

        tmpRect := FShadeRect;

        { decrease the size of the shaded rect by 1 so that it does not obscure }
        { the gridlines already drawn }
        if (FPrintInfo.ShowGridLines) then
        begin
          { shrink the cell area according to each of the gridlines }
          inc(tmpRect.Left, GridLines.Left);
          inc(tmpRect.Top, GridLines.Top);
          dec(tmpRect.Right, GridLines.Right);
          dec(tmpRect.Bottom, GridLines.Bottom);
        end
        else
        begin
          { there are no gridlines, so shift cell boundaries so adjacently shaded }
          { cells appear seamless (and not with white spaces where the gridlines }
          { would have gone) }
          inc(tmpRect.Right);
          inc(tmpRect.Bottom);
        end;

        { shade the cell }
        FillRect(tmpRect);

        Brush.Style := OldBrushStyle;
        Brush.Color := OldBrushColor;
      end;
    end;
  end;

  { Set the DrawText options }
  { mask off any other options other than alignment from the textjustify field }
  DrawOptions := FDrawInfo.TextJustify and (DT_LEFT or DT_RIGHT or DT_CENTER or DT_VCENTER);

  { set whether to use wordwrapping or not }
  if (coWordWrap in FDrawInfo.Options) then
    DrawOptions := DrawOptions or DT_WORDBREAK
  else
    DrawOptions := DrawOptions or DT_SINGLELINE;

  OldBrushStyle := FDrawInfo.PrintStatus.ActiveCanvas.Brush.Style;
  FDrawInfo.PrintStatus.ActiveCanvas.Brush.Style := bsClear;

  { display the text }
  DrawText(FDrawInfo.PrintStatus.ActiveCanvas.Handle, @FDrawInfo.CellText[1],
           Length(FDrawInfo.CellText), FDrawInfo.CellRect,
           DrawOptions);

  with FDrawInfo.PrintStatus.ActiveCanvas.Brush do
    Style := oldBrushStyle;
end;

procedure TPrintedReport.BuildColumnWidthsList;
{ get the column width of each column of each grid }
{ then trigger an OnGetColWidth event, allowing the user to }
{ modify these values without affecting the grid }
{ because, the grid may be an onscreen grid }
var
  GridNo, iCol :integer;
  ColWidth :longint;
  tmpCollect :TList;
  tmpGrid :TStringGrid;
  tmpReportColumn :TReportColumn;
  PrintGetColWidthEvent :PRN_TPrintGetColWidthEvent;
begin
 { for each grid, build placeholding maxwidth object for each column }
  for GridNo := 0 to FGrids.Count - 1 do
  begin
   { create a list to hold all of the columns maximum widths }
    tmpCollect := TList.create;
    tmpGrid := TStringGrid(FGrids.Items[GridNo]);

   { create a blank colrec for each column in the current grid }
    for iCol := 0 to tmpGrid.ColCount - 1 do
    begin
      tmpReportColumn := TReportColumn.create;
      tmpCollect.Add(tmpReportColumn);
    end;

    { add the list of column maxwidths to the list of TReportColumns }
    FPageSpreads.Add(tmpCollect);
  end;

 { For each col, scan through each grid and store the maximum col width }
  for GridNo := 0 to FGrids.Count - 1 do
  begin
    tmpGrid := TStringGrid(FGrids.Items[GridNo]);
    tmpCollect := TList(FPageSpreads.Items[GridNo]);

    { for each column }
    for icol := 0 to tmpGrid.ColCount - 1 do
    begin
      { get the column record for this grids column }
      tmpReportColumn := tmpCollect.Items[icol];

      { take a copy of the column width allowing it to be }
      { modified by the user and then scaled (for fit to page) }
      { without the chance of it affecting the grid, if for instance }
      { it is on screen }
      ColWidth := tmpGrid.ColWidths[icol];

      { trigger an OnPrinterGetColWidth event for the current column }
      { allowing the unscaled column width to be modified }
      PrintGetColWidthEvent := PrintInfo.OnPrintGetColWidth;
      if assigned(PrintGetColWidthEvent) then
        PrintGetColWidthEvent(self, GridNo, iCol, ColWidth);

      { assign the processed column width to the column record }
      tmpReportColumn.MaxWidth := ColWidth;
    end;

    { now for each grid, add a placeholder string for each of the max spreads }
    FMaxSpread.Add('1');
  end;
end;

procedure TPrintedReport.setDefaultDrawInfo(DrawInfo :PRN_TDrawInfoStruct);
begin
  FDefaultDrawInfo := DrawInfo;

  { make sure the active canvas is always assigned (incase whole record was assigned) }
  FDefaultDrawInfo.PrintStatus.ActiveCanvas := Printer.Canvas;
end;

procedure TPrintedReport.SetColumnScope;
{ sets the start and end column variables which looks up how many of }
{ the unprinted columns remaining, can be fit onto the current pagespread }
var
  ThisGrid :TStringGrid;
  TmpCollect :TList;
begin
  ThisGrid := TStringGrid(FGrids.Items[FiGrid]);
  TmpCollect := TList(FPageSpreads.Items[FiGrid]);

 { find the first column which is on the current PageSpread }
  FStartCol := ThisGrid.FixedCols;
  while ((TReportColumn(TmpCollect.Items[FStartCol]).PageSpread < FCurrentSpread) and
         (FStartCol < ThisGrid.ColCount - 1)) do
    inc(FStartCol);

 { if the last column in the grid already appeared on a previous page spread }
 { then set end col to -1. This will omit the grid, but still process the printing }
 { so that a vertical space corresponding to its height is left on pagespreads }
 if (TReportColumn(TmpCollect.Items[ThisGrid.ColCount - 1]).PageSpread < FCurrentSpread) then
     FEndCol := -1
  else
  begin
    { otherwise find the last column record which can be included on this PageSpread }
    FEndCol := FCurrentGrid.ColCount - 1;
    while ((FEndCol > 1) and
           (TReportColumn(TmpCollect.Items[FEndCol]).PageSpread > FCurrentSpread)) do
      dec(FEndCol);
  end;
end;

procedure TPrintedReport.SetPrinterFont(aScreenFont :TFont);
{ assuming the admin variables have been set, this will take any font as a parameter }
{ and scale it according to the printers resolution, so that it appears in WYSIWYG }
var
  TextInch :real;
begin
  with FDrawInfo.PrintStatus do
  begin
    { Calculate the height of the reportviewers font in inches }
     if (FPrintInfo.PixelsPerInch <> 0) then
       TextInch := aScreenFont.Height / FPrintInfo.PixelsPerInch
     else
       raise Exception.create('Blocked attempt to scale Printer font using an unspecified screen resolution.');

     { assign the font to the Printer, and scale its height accordingly }
     with ActiveCanvas do
     begin
       Font := aScreenFont;
       Font.Height := Trunc(TextInch * FPixPerInY * FScaling_Factor);
     end;
  end;
end;

procedure TPrintedReport.GetPrinterResolutions;
{ get horizontal and vertical Printer resolutions }
begin
  with FDrawInfo.PrintStatus.ActiveCanvas do
  begin
    FPixPerInX := Trunc(GetDeviceCaps(Handle, LOGPIXELSX));
    FPixPerInY := Trunc(GetDeviceCaps(Handle, LOGPIXELSY));
  end;
end;

procedure TPrintedReport.DrawGridLines(aRect :TRect);
var
  OldPenColor :TColor;
  OldPenWidth :longint;
  OldPenStyle :TPenStyle;
  lp :integer;
begin
  with FDrawInfo do
    with PrintStatus do
      with ActiveCanvas do
      begin
        { preserve the pen settings }
        with Pen do
        begin
          OldPenColor := Color;
          OldPenWidth := Width;
          OldPenStyle := Style;
        end;

        Pen.Color := clBlack;
        Pen.Style := psSolid;

        { Top Line }
        if (GridLines.Top > 0) then
        begin
          Pen.Width := 1;

          for lp := 0 to GridLines.Top - 1 do
          begin
            MoveTo(aRect.Left, aRect.Top + lp);
            LineTo(aRect.Right, aRect.Top + lp);
          end;
        end;

        { Left Line }
        if (GridLines.Left > 0) then
        begin
          Pen.Width := 1;

          for lp := 0 to GridLines.Left - 1 do
          begin
            MoveTo(aRect.Left + lp, aRect.Top);
            LineTo(aRect.Left + lp, aRect.Bottom);
          end;
        end;

        { Bottom Line }
        if (GridLines.Bottom > 0) then
        begin
          Pen.Width := 1;

          for lp := 0 to GridLines.Bottom - 1 do
          begin
            MoveTo(aRect.Left, aRect.Bottom - lp);
            LineTo(aRect.Right, aRect.Bottom - lp);
          end;
        end;

        { Right Line }
        if (GridLines.Right > 0) then
        begin
          Pen.Width := 1;

          for lp := 0 to GridLines.Right - 1 do
          begin
            MoveTo(aRect.Right - lp, aRect.Top);
            LineTo(aRect.Right - lp, aRect.Bottom);
          end
        end;

        { restore the pen settings }
        with Pen do
        begin
          Color := OldPenColor;
          Width := OldPenWidth;
          Style := OldPenStyle;
        end;
      end;
end;

{WBF136.PAC - START}
procedure TPrintedReport.InstallHeaderFont;
begin
  { validate the font index specified to use for the description and header }
  ValidateHeaderFontIndex(FPrintInfo);

  { save the current printer font into FOldFont }
  if (assigned(FOldFont)) then
    FOldFont.Assign(Printer.Canvas.Font);

  { Set the current font, to the font of the canvas of the grid at the position }
  { specified in PrintInfo.HeaderFontIndex }
  SetPrinterFont(TStringGrid(FGrids.Items[PrintInfo.HeaderFontIndex]).Canvas.Font);
end;

procedure TPrintedReport.UnInstallHeaderFont;
begin
  { reinstall the old font, before InstallHeaderFont was called }
  if assigned(FOldFont) then
    Printer.Canvas.Font.Assign(FOldFont);
end;

procedure TPrintedReport.ValidateHeaderFontIndex(var PrintInfo :PRN_TPrintInfoStruct);
begin
  with PrintInfo do
  begin
    { make sure headerfont index is an index to a valid grid in the collection }
    if (HeaderFontIndex >= FGrids.Count) then
      HeaderFontIndex := FGrids.Count - 1
    else
    if (HeaderFontIndex < 0) then
      HeaderFontIndex := 0;
  end;
end;
{WBF136.PAC - END}

end.
