{--------------------------------------------------------------------------------}
{                                                                                }
{   UNIT : Printing Types                                                        }
{                                                                                }
{   DATE        : 25/04/97                                                       }
{   AUTHOR      : P.Chapman                                                      }
{   STATUS      : Development                                                    }
{                                                                                }
{--------------------------------------------------------------------------------}
{  Modifications:                                                                }
{                                                                                }
{  W1 02/06/97.PAC - Moved some exporting enumurated types in from RT_TYPE       }
{  W7 21/06/97.PAC - Implemented Print Manager object                            }
{  W8 21/06/97.PAC - Changed OnExportEvent to supply a structure                 }
{  WBF136      PAC - Added PrintInfo.HeaderFontIndex. Setting this to 4 for      }
{                    would tell the printed report to use the font of the fourth }
{                    stringgrid, when printing headers and descriptions.         }
{--------------------------------------------------------------------------------}

unit PRN_Type;

interface

uses
  Graphics,
  Grids,
  WinTypes,
  Classes;

const
  clVeryLightGrey = $00EDEDEB;
  CarriageReturn = #10#13;
  clPrintCellShadeColor = clYellow;

type
  {------------------}
  { ENUMERATED TYPES }
  {------------------}

  PRN_TOptionsDialogContext = (dcExportOptions, dcPrintOptions);
  PRN_TCellDrawingOption    = (coShaded, coWordWrap);
  PRN_TCellType             = (ctString, ctReal);
  PRN_TCellDrawingOptions   = set of PRN_TCellDrawingOption;

  { used to record the method of exporting a report }
  PRN_TExportType           = (etFixed, etDelimited, etExcel);

  {------------}
  { STRUCTURES }
  {------------}

  { the PrintStatus structure holds info on the current printing status }
  { This is accessible in the OnPrintCell event, through the DrawInfo structure }
  { but its fields should not be changed }
  PRN_TPrintStatusStruct = record
    ActiveCanvas :TCanvas;           { pointer to the canvas to send all output to }
    GridIndex :Integer;              { index to the current grid }
    Row, Col :LongInt;               { current cell to be drawn }
    CurrentGrid :TStringGrid;        { pointer to the current grid }
  end;

  { The structure which holds data on how the current cell is to be printed }
  { This is available in the OnPrintCell Event, and changes can be made to its fields }
  { and the currently printing cell will be printed accordingly }
  { Fields within the PrintStatus record should be treated as read only }
  PRN_TDrawInfoStruct = record
    PrintStatus :PRN_TPrintStatusStruct; { holds details of the progress of the printing }
    CellRect :TRect;                     { bounding box to draw cell into }
    TextJustify :integer;                { justification of text, as in DT_LEFT etc.. }
    Options :PRN_TCellDrawingOptions;    { set of options which affect the printed cells appearance }
    CellText :string;                    { the cell text which is about to be printed }
    DefaultDrawing :Boolean;             { whether to follow a printcell event up with the generic drawcell }
    ShadeColor :TColor;                  { the color used to shade the fixed rows / cols }
    GridLines :TRect;                    { holds the pen width for each gridline around a cell }
  end;

  { The structure which holds info on the current cell being exported }
  PRN_TExportInfoStruct = record
    Grid :TStringGrid;
    iRow, iCol :longint;
    CellText :string;
    TextJustify :longint;
    CellType :PRN_TCellType;
    ExportType :PRN_TExportType;
  end;

  {-------------------}
  { EVENT DEFINITIONS }
  {-------------------}

  { event type used by the printed report and flexi report types }
  PRN_TPrintCellEvent      = procedure(Sender :TObject; var Attributes :PRN_TDrawInfoStruct) of object;

  PRN_TPrintNewPageEvent = procedure (Sender :TObject; CurrentPage, TotalPages :integer;
                                var Xcursor, YCursor :longint) of object;

  PRN_TPrintNewGridEvent = procedure (Sender :TObject; NewGridIndex :integer) of object;

  PRN_TPrintGetColWidthEvent = procedure (Sender :TObject;
                                           GridIndex, iCol :integer;
                                           var ColWidth :longint);

  PRN_TExportGetColWidthEvent = procedure (Sender :TObject; iCol :integer; var ColWidth :integer) of object;

  PRN_TExportCellEvent = procedure(Sender :TObject; var ExportInfo :PRN_TExportInfoStruct) of object;

  PRN_TPrintInfoStruct = record
    ShowGridLines :Boolean;
    ShowDescription :Boolean;
    ShowHeader :Boolean;
    FitToPage :Boolean;
    PixelsPerInch :integer;
    GridSpacing :integer;
    OmitFixedRows :Boolean;

    Description :TStringList;
    Header :string;
    HeaderFontIndex :integer;
    StartNewPage :Boolean;

    OnPrintNewPage :PRN_TPrintNewPageEvent;
    OnPrintCell :PRN_TPrintCellEvent;
    OnPrintNewGrid :PRN_TPrintNewGridEvent;
    OnPrintGetColWidth :PRN_TPrintGetColWidthEvent;
  end;

{  PRN_TPrintManager = class(TObject)
  private
  public
    procedure PRN_InitDrawInfoStruct(var DrawInfo :PRN_TDrawInfoStruct);
    procedure PRN_InitPrintInfoStruct(var PrintInfo :PRN_TPrintInfoStruct);
  end; }

procedure PRN_InitDrawInfoStruct(var DrawInfo :PRN_TDrawInfoStruct);
procedure PRN_InitPrintInfoStruct(var PrintInfo :PRN_TPrintInfoStruct);

implementation

uses
  Forms;

procedure PRN_InitDrawInfoStruct(var DrawInfo :PRN_TDrawInfoStruct);
{ sets the standard default values for the DrawInfo Structure }
{ these fields are set before each OnPrintCell event }
{ but it is assumed that the PrintStatus subrecord has been prefilled }
begin
  { blank out the drawinfo record }
  FillChar(DrawInfo, SizeOf(DrawInfo), #0);

  { set the default values of the drawinfo }
  with DrawInfo do
  begin
    TextJustify := DT_LEFT or DT_VCENTER;;
    Options := [];
    DefaultDrawing := True;
    ShadeColor := clPrintCellShadeColor;
    CellRect := Rect(0, 0, 0, 0);
    CellText := '';
    GridLines := Rect(1, 1, 1, 1);

    { the printstatus structure contains invalid data until it actually is populated }
    { and recieved in the OnPrintCell handler }
    with PrintStatus do
    begin
      Row := -1;
      Col := -1;
      GridIndex := -1;
      CurrentGrid := nil;
      ActiveCanvas := nil;
    end;
  end;
end;

procedure PRN_InitPrintInfoStruct(var PrintInfo :PRN_TPrintInfoStruct);
begin
  { blank out the Printinfo structure }
  FillChar(PrintInfo, SizeOf(PrintInfo), #0);

  { set the default values for the printinfo structure }
  with PrintInfo do
  begin
    OnPrintNewGrid := nil;
    OnPrintCell := nil;
    OnPrintNewPage := nil;
    OnPrintGetColWidth := nil;
    Description := nil;

    Header := '';                            
    HeaderFontIndex := 0;
    GridSpacing := 20;
    PixelsPerInch := Screen.PixelsPerInch;

    ShowGridLines := True;
    ShowDescription := True;
    ShowHeader := True;
    FitToPage := False;
    OmitFixedRows := False;
    StartNewPage := False;
  end;
end;

end.
