{------------------------------------------------------------------------------}
{           UNIT :   XL_TYPES.PAS                                              }
{                                                                              }
{         AUTHOR :   P.CHAPMAN                                                 }
{                                                                              }
{           DATE :   18.10.96                                                  }
{                                                                              }
{------------------------------------------------------------------------------}
{    DESCRIPTION :   Defines the EXCEL export object, allowing a grid          }
{                    to be exported from the paymodeller into MS EXCEL.        }
{                    Creates an Excel single sheet in BIFF 5/7 format.         }
{------------------------------------------------------------------------------}
{ CLASSES DEFINED                                                              }
{                                                                              }
{   TIndex, TFontRec, TFontStore, TXF, TXFStore, TDimensions, TCell,           }
{   TDoubleRec, TWindow1, TWindow2, TExcelSheet, TFontStore, TXFStore,         }
{   TCellStore, TStyleStore, TRowStore, TRow, TStyle, TFont, TCell             }
{   TLabelCell, TNumberCell                                                    }
{                                                                              }
{------------------------------------------------------------------------------}
{ DEVELOPMENT HISTORY                                                          }
{                                                                              }
{  17/10/96                                                                    }
{     1. Wrote procedures to format data types to 8087 format                  }
{     2. Started coding WriteRec procedures to output specific records.        }
{     3. Introduced record ID constants, and WriteRecHeader procedure.         }
{     4. Localised record length constants and variables                       }
{  18/10/96                                                                    }
{     1. more WriteRecord procedures.                                          }
{     2. TFormatStore using a StringList.                                      }
{  23/10/96                                                                    }
{     1. TFontstore to hold unique registered fonts.                           }
{  25/10/96                                                                    }
{     1. Started encapsulating some simple records into properties.            }
{     2. Implemented TXFStore to hold unique registered XF's.                  }
{  27/10/96                                                                    }
{     1. RegisterMinimumStyleSet procedure.                                    }
{  21/11/96                                                                    }
{     1. Allow duplicates in minimum record sets for fonts, style, formats     }
{  23/01/97                                                                    }
{     PAC 1. Implemented TDBCELL class                                         }
{  24/01/97                                                                    }
{     PAC 1. Implemented TRowStore.Sort, TCellStore.Sort.                      }
{     PAC 2. Register minimum sets on create, instead of explicit calls        }
{                                                                              }
{ W8     22/06/97 PAC - Added new minimum XF which holds spec for a bold cell  }
{                     - Added TColInfoStore, TColInfo objects                  }
{                     - Added WriteColInfo procedure                           }
{ WBF131 22/06/97 PAC - Cells were being added using CellStore.Add instead of  }
{                       AddCell procedure. Admin info was becoming inconsistent }
{------------------------------------------------------------------------------}

unit XL_Types;

interface

uses
  Messages, SysUtils, Classes,
  Graphics, Controls, Forms, Dialogs,
  StdCtrls, WinProcs, Grids, BigList;

type
{ ------------------------------ BASE TYPES ----------------------------- }

 { Typedef LongWord to be DWORD, for readability }
  LongWord = LongInt;

 { xlCellType is used to record which derivative of TCell is being used. }
 { stored in base class TCell }
  xlCellType = (ctLabel, ctNumber, ctBlank);

  TXLValidation = (xvOK, xvTooManyRows, xvTooManyCols, xvSheetTooLarge);

  XL_TTextStyle = (tsNormal, tsBold);
  
{ ------------------------------- RECORDS ------------------------------- }

  { TDOUBLEREC union record. Allows a Double to be also viewed as individual bytes }
  TDoubleRec = record
    case Boolean of
     True: (DoubleView :Double);
    False: (ByteView :array [1..8] of byte);
  end;

  { INDEX record. Stores first/last row used, and an array of stream offsets to each DBCELL }
  TIndex = record
    FirstRefdRow :word;
    LastRow :word;
    DBCellOffsets : TBigList;
  end;

  { SETUP record. Used to store settings for the Page Setup }
  TSetup = record
    PaperSize :word;            { iPapersize    Paper size }
    Scale :word;                { iScale        Scaling factor }
    PageStart :word;            { iPageStart    Starting page number }
    FitWidth :word;             { iFitWidth     Fit to width: number of pages }
    FitHeight :word;            { iFitHeight    Fit to height: number of pages }
    Attributes :word;           { grbit         Options flags (see below) }
    Resolution :word;           { iRes          Horizontal Resolution }
    VResolution :word;          { iVRres        Vertical Resolution }
    HdrMargin :longword;        { inumHdr       Header margin (inches) }
    FtrMargin :longword;        { inumFtr       Footer margin (inches) }
    Copies :word;               { iCopies       Number of copies }
  end;

  { DEFAULTROWHEIGHT record. Stores the row height for undefined rows }
  TDefaultRowHeight = record
    Attributes :word;           { grbit         bit 0:fUnsynched, 1:fDyZero, 2:fExAsc, 3:fExDsc }
    Height :word;               { miyRw         Default row height (in 1/20ths pt) }
  end;

  { WINDOW1 record. Used to hold information on the window at a workbook level }
  TWindow1 = record
    XPos :word;                 { xWn           X-position of window }
    YPos :word;                 { yWn           Y-position of window }
    Width :word;                { dX            width of window }
    Height :word;               { dX            height of window }
    Attributes :word;           { grbit         windows attributes }
    CurrentTab :word;           { itabCur       current tab }
    FirstTab :word;             { itabFirst     first visible tab }
    NoOfSelectedTabs :word;     { ctabSel       number of tabs selected }
    TabRatio :word;             { wTabRatio     width ratio (Tab : Hscrollbar) }
  end;

  { WINDOW2 record. Used to hold information on the window at a worksheet level }
  TWindow2 = record
    Attributes :word;           { grbit         windows attributes }
    TopRow :word;               { rwtop         topmost visible row }
    LeftCol :word;              { colLeft       leftmost visible column }
    GridColor :word;            { rgbHdr        color of headings/gridlines }
  end;

  { DIMENSION record holds the extreme cell boundaries }
  TDimensions = record
    FirstRow :word;
    LastRow :word;
    FirstCol :word;
    LastCol :word;
  end;

{ ------------------------------- CLASSES ------------------------------- }

  { CELL record. }
  { TCell is an abstract class with TLabelCell, TNumberCell etc. derived from it  }
  TCell = class
    Row, Col :word;
    XFIndex  :word;
    CellType :xlCellType;
  end;

  { NUMBERCELL record }
  TNumberCell = class(TCell)
    Content  :double;
  end;

  { LABELCELL class }
  TLabelCell = class(TCell)
    Content  :string;
  end;

{ FORMAT: Used to hold a format registered in the FORMATSTORE }
  TFormat = class(TObject)
  private
    iFormat :word;
    FormatStr :string;
  public
    constructor create;
  end;

{ ROW: Used to hold attributes of a defined row }
  TRow = class(TObject)
  private
    RowNo :word;                { rw            row number (zero based) }
    FirstCol :word;             { colMic        first defined column in row }
    LastCol1 :word;             { colMac        last defined column in row + 1 }
    RowHeight :word;            { miyRw         row height }
    Attributes :word;           { grbit         iOutLevel, fCollapsed, fDyZero, fUnsynched, fGhostDirty }
    XFIndex :word;              { ixfe          if fGhostDirty then index to XF. Also fExAsc,fExDsc }
  public
    constructor Create;
  end;

  { DBCELL class. }
  { Used to specify offsets backwards from the DBCELL to the start of }
  { the second ROW in the block. One DBCELL required for every block of 32 ROWs }
  { Also specifies an array of offsets from the start of the first row to }
  { the start of every CELL in the current block of ROW's }
  TDBCell = class(TObject)
  private
    FBlockOffset :integer;
    FCellOffsets :TBigList;
  public
    constructor create;
    destructor destroy; override;
    procedure ResetOffsets;
    procedure SetOffset(cellno, offset :integer);
    procedure BuildCellOffsetList(NoOfCells :integer);
  end;

  TOffset = class(TObject)
  private
{    FCellNumber :integer; }
    FOffset :integer;
  public
  end;

 { FONT class. Used to hold a font registered in the sheet }
  TFontRec = class(TObject)
  private
    Height :word;               { dyHeight      font height, (1/20th pts) }
    Attributes :word;           { grbit         attributes }
    Color :word;                { icv           color index }
    Weight :word;               { bls           weight (100..1000) }
    Script :word;               { sss           script style }
    Underline :word;            { uls           underline style }
    Family :byte;               { bFamily       API Font family }
    CharSet :byte;              { bCharset      API Font character set }
    Name: string;               { Name          Windows font name }
  public
    constructor create;
  end;

  { STYLE class. Used to hold a registered style }
  TStyle = class
  private
    XFindex :word;              { index to XF record, only uses lowest 12 bits }
    IsBuiltIn :ByteBool;        { True if a built in style }
    BuiltIn :byte;              { If IsBuiltIn then index of Built In style  }
    Level :byte;                { for a built in style }
    Name :string;               { If not IsBuiltIn, then contsinas user defined style name }
  public
    constructor create;
  end;

  { XF class. Used to store a registered extended format (XF record) }
  TXF = class
  private
  { XFFont : word; }         { index to the Font record }
    XFFormat :byte;          { index to format record }
    XFInfo :byte;            { locked, hidden, XF type, XF parent index }
    XFTextInfo :word;        { alignment, rotation, attributes }
    FillColors :word;        { index into palette for cell fill pattern color }
    BorderInfo :word;        { FillPattern, Bottomborder line style }
    BorderInfo2 :word;       { Top/left/right border style, TopBorder color }
    FillColors2 :word;       { left/right border colors }
  public
    constructor Create;
  end;

  { COLINFO: Used to store column width, formatting and options }
  TColInfo = class(TObject)
  private
  public
    Column :word;
    ColWidth :word;
    XFIndex :word;
    Options :word;
    constructor create;
  end;

{ -------------------------- CONTAINER CLASSES -------------------------- }

  { CELLSTORE: Used to hold all registered cells }
   TCellStore = class(TBigList)
   private
   public
     constructor create;
     procedure CellStoreOnCompare(ItemToInsert, Key : TObject;
       Var Index : TCompareResult; Var PartialResult : Boolean);
   end;

  { FORMATSTORE: Used to hold all defined formats }
  TFormatStore = class(TBigList)
  private
    function FindDuplicate(ThisFormat :TFormat) :Integer;
  public
    constructor Create;
    function RegisterFormat(aFormat :TFormat) :integer;
  end;

  { ROWSTORE: Used to hold all defined ROW's }
  TRowStore = class(TBigList)
  private
  public
    constructor Create;
    function RegisterRow(aRow :TRow) :integer;
    procedure RowStoreOnCompare(ItemToInsert, Key : TObject;
             Var Index : TCompareResult; Var PartialResult : Boolean);
  end;

  { FONTSTORE: Used to hold all registered fonts }
  TFontStore = class(TBigList)
  private
    function FindDuplicate(ThisFont :TFontRec) :Integer;
  public
    constructor Create;
    function RegisterFont(aFont :TFontRec) :Integer;
  end;

  { STYLESTORE: Used to hold all unique registered styles }
  TStyleStore = class(TBigList)
  private
    function FindDuplicate(ThisStyle :TStyle) :Integer;
  public
    constructor Create;
    function RegisterStyle(aStyle :TStyle) :Integer;
  end;

  { XFSTORE: Used to hold list of all registered XF's }
  TXFStore = class(TBigList)
  private
    function FindDuplicate(ThisXF :TXF) :Integer;
  public
    constructor Create; 
    function RegisterXF(anXF :TXF) :Integer;
  end;

  TColInfoStore = class(TList)
  private
  public
    procedure AddColInfo(startcol, endcol, colwidth :word);
    function FindColInfo(index :integer) :TColInfo;
  end;

{ ########################################## }
{   CLASS: TExcelSheet - SpreadSheet object  }
{ ########################################## }
  TExcelSheet = class(TObject)
  private
  { Output streams }
    outfile  : TFileStream;                             { output stream for final excel file }
    inFile   : TMemoryStream;                           { input stream for reading in excel file }

  { internal book-keeping fields }
    FNoOfCells, FNoOfRows :longint;
    FDBCell :TDBCell;                                   { DBCELL }
    FStreamOffset :longint;                             { Byte counter (but not necc. since file started) }
    FBlockOffset :longint;                              { Offset of current row block for DBCELL }
    FCellOffset :longint;                               { Offset of current cell used in DBCELL }
    FMaxCols :word;                                     { maximum number of cols (limited by XL itself) }
    FMaxRows :word;                                     { maximum number of rows (limited by XL itself) }

  { Property-records fields }
    FWriteAccess :string;                               { WRITEACCESS }
    FLeftMargin :double;                                { LEFTMARGIN }
    FRightMargin :double;                               { RIGHTMARGIN }
    FTopMargin :double;                                 { TOPMARGIN }
    FBottomMargin :double;                              { BOTTOMMARGIN }
    FHeader :string;                                    { HEADER }
    FBackup :Boolean;                                   { BACKUP }
    FCountry :Word;                                     { COUNTRY }
    FWindowProtect :Boolean;                            { WINDOWPROTECT }
    FFooter :string;                                    { FOOTER }
    FIteration :Boolean;                                { ITERATION }
    FDelta :word;                                       { DELTA }
    FCodePage :word;                                    { CODEPAGE }
    FCalcCount :word;                                   { CALCCOUNT }
    FCalcMode :Boolean;                                 { CALCMODE }
    FProtect :Boolean;                                  { PROTECT }
    FRefMode :Boolean;                                  { REFMODE }
    FSaveRecalc :Boolean;                               { SAVERECALC }
    FFileName :string;
    FDefColWidth :word;                                 { DEFCOLWIDTH }
    FDefaultFormat :word;                               { DEFAULTFORMAT }
    FVCenter :Boolean;                                  { VCENTER }
    FHCenter :Boolean;                                  { HCENTER }
    FPrintHeaders :Boolean;                             { PRINTRWCOL }
    FWSBool :word;                                      { WSBOOL }
    FGuts :word;                                        { GUTS }
    FPrintGridLines :Boolean;                           { PRINTGRIDLINES }

    FTextStyle :XL_TTextStyle;                          { bold / normal }

  { Records }
    Index               :TIndex;                        { INDEX }
{    DefaultRowHeight    :TDefaultRowHeight;             { DEFAULTROWHEIGHT }
    PageSetup           :TSetup;                        { SETUP }
    Dimensions          :TDimensions;                   { DIMENSIONS }
    BookWindow          :TWindow1;                      { WINDOW1 }
    SheetWindow         :TWindow2;                      { WINDOW2 }

  { Container classes }
    FormatStore         :TFormatStore;                  { FORMAT's }
    StyleStore          :TStyleStore;                   { STYLE's }
    FontStore           :TFontStore;                    { FONT's }
    RowStore            :TRowStore;                     { ROW's }
    XFStore             :TXFStore;                      { XF's }
    CellStore           :TCellStore;                    { CELL'S }
    ColInfoStore        :TColInfoStore;                 { COLINFO'S }

  protected
  { adds a generic TCell to the CellStore }
    procedure AddCell(NewCell :TCell);

  { methods used to add mandatory records for a skeleton sheet }
    procedure RegisterMinimumFontSet;
    procedure RegisterMinimumStyleSet;
    procedure RegisterMinimumFormatSet;
    procedure RegisterMinimumXFSet;

  { Output methods }
    procedure WriteByte(x: Byte);                               { writes a byte to the stream }
    procedure WriteBytes(a, b: Byte);                           { writes 2 bytes }
    procedure WriteWord(x: word);                               { writes a 2 byte word }
    procedure WriteLongWord(x: longint);                        { writes a  4 byte longword }
    procedure WriteText(s: string; WordLen: Boolean);           { writes a string to stream }
    procedure WriteIEEE(s: double);                             { Writes an 8 byte IEEE floating point number }
    procedure WriteRecHeader(recid, reclen :word);              { Writes an xlRecords header data }
    procedure WriteBOF;
    procedure WriteWriteAccess;
    procedure WriteIndex;
    procedure WriteCodePage;
    procedure WriteCalcCount;
    procedure WriteCalcMode;
    procedure WriteProtect;
    procedure WriteWindowProtect;
    procedure WriteBackup;
    procedure WriteSaveRecalc;
    procedure WriteGuts;
    procedure WriteWSBool;
    procedure WriteRefMode;
    procedure WriteDelta;
    procedure WriteIteration;
    procedure WritePrintGridLines;
    procedure WritePrintHeaders;
    procedure WriteGridSet;
    procedure WriteDefRowHeight;
    procedure WriteFont(NewFont :TFontRec);
    procedure WriteHeader;
    procedure WriteFooter;
    procedure WriteHCenter;
    procedure WriteVCenter;
    procedure WriteLeftMargin;
    procedure WriteRightMargin;
    procedure WriteTopMargin;
    procedure WriteBottomMargin;
    procedure WriteSetup;
    procedure WriteCountry;
    procedure WriteFormat(aFormat :TFormat);
    procedure WriteXF(xfrec :TXF);
    procedure WriteStyle(aStyle :TStyle);
    procedure WriteBuiltInFormatCount;
    procedure WriteName;
    procedure WriteWinProt;
    { procedure WritePalette }
    procedure WriteSelection;
    procedure WriteRGCW;
    procedure WriteDimensions;
    procedure WriteRow(aRow :TRow);
    procedure WriteDefColWidth;
    procedure WriteLabel(aLabel :TLabelCell);
    procedure WriteNumber(aNumberCell :TNumberCell);
    procedure WriteCell(aCell :TCell);
    procedure WriteDBCell(aDBCell :TDBCell);
    { procedure WriteHideObj; }
    procedure WriteWindow1;
    procedure WriteWindow2;
    { procedure WriteSCL }
    { procedure WriteSelection }
    { procedure WriteGCWCols }
    procedure WriteFNGroupCount;
    procedure WriteEOF;

    procedure WriteColInfo(ColInfo :TColInfo);

  { Input methods }
    procedure ReadFromFile(filename :string);
    procedure FeedIntoGrid(aGrid :TStringGrid);

    function ReadByte :Byte;
    function ReadWord :word;
    function ReadLongWord :longint;
    function ReadIEEE :double;
    function ReadText(WordLen: Boolean) :string;
    procedure ReadRecHeader;
    procedure ReadAndDiscard(noOfBytes :word);
    procedure ReadLabel;
    procedure ReadNumber;

  { Helper methods }
    procedure SetBit(var InByte, iBit :byte; State: Boolean);   { Set/Reset a bit in a byte }
    procedure Watch(x:string);                                  { DEBUG: Watch }

  { Property access methods }
    procedure SetWriteAccess(newWriteAccess :string);
    procedure SetCodePage(newCodePage :word);
    function getColumnWidth(index :integer) :word;
    procedure setColumnWidth(index :integer; ColumnWidth :word);

  public
  { Properties }
    constructor Create;
    destructor Destroy; override;

  { ############## }
  { # PROPERTIES # }
  { ############## }

    property MaxCols :word                read FMaxCols;          { Read-only }
    property MaxRows :word                read FMaxRows;          { Read-only }
    property NoOfCells :LongInt           read FNoOfCells;        { Read-only }
    property NoOfRows  :LongInt           read FNoOfRows;         { Read-only }

    property WriteAccess :string          read FWriteAccess       write SetWriteAccess;
    property CodePage :word               read FCodePage          write SetCodePage;
    property Iteration :Boolean           read FIteration         write FIteration;
    property Delta :word                  read FDelta             write FDelta;
    property Backup :boolean              read FBackup            write FBackup;
    property Country :word                read FCountry           write FCountry;
    property Header :string               read FHeader            write FHeader;
    property Footer :string               read FFooter            write FFooter;
    property CalcCount :word              read FCalcCount         write FCalcCount;
    property DefColWidth :word            read FDefColWidth       write FDefColWidth;
    property FileName :string             read FFileName          write FFileName;
    property TextStyle :XL_TTextStyle     read FTextStyle         write FTextStyle;
    property DefaultFormat :word          read FDefaultFormat     write FDefaultFormat;
    property HCenter :Boolean             read FHCenter           write FHCenter;
    property VCenter :Boolean             read FVCenter           write FVCenter;
    property TopMargin :double            read FTopMargin         write FTopMargin;
    property BottomMargin :double         read FBottomMargin      write FBottomMargin;
    property LeftMargin :double           read FLeftMargin        write FLeftMargin;
    property RightMargin :double          read FRightMargin       write FRightMargin;
    property PrintHeaders :Boolean        read FPrintHeaders      write FPrintHeaders;
    property CalcMode :Boolean            read FCalcMode          write FCalcMode;
    property Protect :Boolean             read FProtect           write FProtect;
    property WindowProtect :Boolean       read FWindowProtect     write FWindowProtect;
    property SaveRecalc :Boolean          read FSaveRecalc        write FSaveRecalc;
    property RefMode :Boolean             read FRefMode           write FRefMode;
    property WSBool :word                 read FWSBool            write FWSBool;
    property Guts :word                   read FGuts              write FGuts;
    property DBCell :TDBCell              read FDBCell            write FDBCell;
    property PrintGridLines :Boolean      read FPrintGridLines    write FPrintGridLines;
    property ColumnWidths[index :integer] :word read getColumnWidth write setColumnWidth;

  { ########### }
  { # METHODS # }
  { ########### }

  { Validation methods }
    function ValidateSheetSize(rows, cols :longint) :TXLValidation;

  { Input methods }
    procedure LoadFromFile(aStringGrid :TStringGrid; filename :string);

  { Output methods }
    procedure WriteToFile;

  { Interface methods }
    procedure AddText(iRow, iCol :Integer; str: string);
    procedure AddNumber(iRow, iCol :Integer; num: double);
  end;

{ *** END OF CLASS DEFINITIONS }
{ *** END OF CLASS DEFINITIONS }
{ *** END OF CLASS DEFINITIONS }

implementation

uses
  xl_const;

function GP_Real2Str(ThisReal : real; Decimals : Byte):string;         { This function turns a real }
var
  thestring : string;                                                { into an extra long string.  }
begin
   Str(ThisReal:11:decimals,thestring);
   GP_Real2Str := thestring;
end;

{ ###################################### }
{   EXCELSHEET PROPERTY ACCESS METHODS   }
{ ###################################### }

procedure TExcelSheet.SetWriteAccess(newWriteAccess :string);
{ Set WRITEACCESS Property-Record }
begin
  if (Length(newWriteAccess) > 31) then          { format string to exactly 31 chars long }
    newWriteAccess := copy(newWriteAccess, 1, 31)  { by either truncating }
  else while (Length(newWriteAccess) < 30) do    { or by right-padding with spaces }
    newWriteAccess := newWriteAccess + ' ';
  FWriteAccess := newWriteAccess;
end;

procedure TExcelSheet.SetCodePage(newCodePage :word);
{ Set CODEPAGE Property-Record }
begin
  if (newCodePage = cpAnsi) or
     (newCodePage = cpMac) or
     (newCodePage = cpMultiPlan) then
  FCodePage := newCodePage;
end;

{ ############################## }
{     STREAM OUTPUT METHODS      }
{ ############################## }

procedure TExcelSheet.WriteByte(x: Byte);
{ Writes byte straight to stream }
begin
  outfile.write(x, 1);

  { This would be a good place to increment counters, to track stream offsets }
  inc(FStreamOffset);
  inc(FCellOffset);
  inc(FBlockOffset);
end;

procedure TExcelSheet.WriteBytes(a, b: Byte);
{ swaps bytes and writes to stream }
begin
  WriteByte(b);
  WriteByte(a);
end;

procedure TExcelSheet.WriteWord(x: word);
{ writes the word in little endian format }
begin
  WriteByte(lo(x));
  WriteByte(hi(x));
end;

procedure TExcelSheet.WriteLongWord(x: longint);
{ writes the longword in little endian format }
begin
  WriteWord(loword(x));
  WriteWord(hiword(x));
end;

procedure TExcelSheet.WriteIEEE(s: double);
{ Writes an IEEE 80x87 format floating point number to the stream }
var
  IEEE :TDoubleRec;
  loop :Integer;
begin
 IEEE.doubleview := s;
  for loop := 1 to 8 do
    WriteByte(IEEE.byteview[loop]);
end;

procedure TExcelSheet.WriteText(s: string; WordLen: Boolean);
{ Writes a string to the stream, using a preceding word to store its length }
var
  byteno :byte;
begin
  if WordLen then
    WriteWord(Length(s))          { use word for length }
  else
    WriteByte(Length(s));         { just use a byte }

  for byteno := 1 to length(s) do   { write string contents }
    WriteByte(byte(s[byteno]));
end;

{ ############################## }
{       CLASS CONSTRUCTORS       }
{ ############################## }

constructor TCellStore.create;
{ Creates an instance of the CELLSTORE }
{ Contains instances of derivatives of TCell }
begin
  inherited create;
{  Sorted := True; }
{  OnCompare := CellStoreOnCompare; }
  Duplicates := dupIgnore;
end;

constructor TFontRec.Create;
{ Creates a FontRec used to hold a Font. }
{ Instances contained by: TFontStore }
begin
  inherited create;
end;

constructor TFontStore.Create;
{ Creates a FontStore.  Holds all fonts registered for the TExcelSheet }
{ Contains instances of: TFontRec }
begin
  inherited create;
  Duplicates := dupIgnore;
end;

constructor TFormatStore.Create;
{ Creates a FormatStore.  Holds all formats registered for the TExcelSheet }
{ Contains instances of: TFormat }
begin
  inherited create;
  Duplicates := dupIgnore;
end;

constructor TFormat.Create;
{ Creates a Format object }
{ Instances contained by: TFormatStore }
begin
  inherited create;
end;

constructor TXF.Create;
{ Creates a new XF object, setting various fields to default values }
{ Instances contained by: TXFStore }
begin
  inherited Create;
end;

constructor TXFStore.Create;
{ Creates a new XFStore object. Setting fields to default values }
{ Contains instances of: TXF }
begin
  inherited Create;
  Duplicates := dupIgnore;
end;

constructor TStyleStore.Create;
{ Creates a new StyleStore object. Setting fields to default values }
{ Contains instances of: TStyle }
begin
  inherited Create;
  Duplicates := dupIgnore;
end;

constructor TStyle.Create;
{ Creates a new Style object. Setting fields to default values }
{ Contains instances of: TStyle }
begin
  inherited Create;
end;

constructor TRow.Create;
{ Creates a new Row object. Used to define a row on a sheet }
{ Instances contained by: TRowStore }
begin
  inherited Create;
  RowHeight := $8000;    { default height }
  FirstCol := $ffff;
  LastCol1 := $00;
end;

constructor TDBCell.create;
begin
  FCellOffsets := TBigList.create;
end;

constructor TRowStore.Create;
{ Creates a new RowStore object, to hold row definitions }
{ Contains instances of: TRow }
begin
  inherited Create;
  Duplicates := dupIgnore;
  Sorted := True;
  OnCompare := RowStoreOnCompare;
end;

constructor TExcelSheet.Create;
{ Creates an ExcelSheet }
begin
  inherited Create;

  FMaxRows := 16000;
  FMaxCOls := 16000;

 { create instances of the container classes to hold all the objects in }
  FormatStore := TFormatStore.create;
  StyleStore := TStyleStore.create;
  FontStore := TFontStore.create;
  XFStore := TXFStore.create;
  RowStore := TRowStore.create;
  CellStore := TCellStore.create;
  ColInfoStore := TColInfoStore.create;

 { create one instance of the DBCELL object }
  DBCell := TDBCell.create;

 { set some default properties }
  TextStyle := tsNormal;
  Header := '&A';
  Footer := 'Page &P';

  { set the write access to be the version text }
  WriteAccess :=  xlAuthor;

  DefColWidth := $0008;

  { set the default filename }
  FileName := '\EXCEL.XLS';
  CodePage := cpAnsi;
  LeftMargin := 0;
  RightMargin := 0;
  TopMargin := 0;
  BottomMargin := 0;

  Index.FirstRefdRow := $ffff;
  Dimensions.FirstCol := $ffff;
  Dimensions.FirstRow := $ffff;
  CalcMode := True;
  RefMode := True;
  Delta := $01;
  Saverecalc := True;
  PrintGridLines := True;

{ reset some counter variables }
  FNoOfRows := 0;
  FNoOfCells := 0;

{ set the default values in the WINDOW1 record }
  with BookWindow do
  begin
    XPos := $0060;                 { Horisontal window position }
    YPos := $0024;                 { Vertical window position }
    Width := $3b1c;
    Height := $27b4;
    Attributes := $0038;
    CurrentTab := $0000;
    FirstTab := $0000;
    NoOfSelectedTabs := $0000;
    TabRatio := $0258;
  end;

  { set the defaults values in the WINDOW2 record }
  with SheetWindow do
  begin
    Attributes := $06b6;          { Attributes }
    TopRow := $0000;              { Topmost visible row }
    LeftCol := $0000;             { Leftmost visible column }
    GridColor := $00000000;       { Color of headers/gridLines }
  end;

{ set the default PageSetup attributes. Indicates invalid printer settings/orientation }
  with PageSetup do
  begin
    PaperSize   := $00;
    Scale       := $2c;
    PageStart   := $01;
    FitWidth    := $01;
    FitHeight   := $01;
    Attributes  := $46;
    Resolution  := $7370;
    VResolution := $6e6f;
    HdrMargin   := $e03f;
    FtrMargin   := $e03f;
    Copies      := $00;
  end;
end;

{ ############################## }
{        CLASS DESTRUCTORS       }
{ ############################## }
destructor TDBCell.Destroy;
var
  listloop :integer;
begin
  for listloop := 0 to FCellOffsets.count-1 do
  begin
    TOffset(FCellOffsets.At(0)).Free;
    FCellOffsets.AtDelete(0);
{    FCellOffsets.Pack; }
  end;

  FCellOffsets.Free;
  FCellOffsets := nil;

  inherited destroy;
end;

{ ############################## }
{        OBJECT METHODS          }
{ ############################## }

procedure TExcelSheet.watch(x:string);
{ takes a string, and displays it in a messagebox. Used for debugging }
begin
  MessageDlg('Watch on: ' + x, mtError, [mbOK], 0);
end;

procedure TExcelSheet.WriteRecHeader(recid, reclen :word);
{ a record header is 4 bytes long. Specifies the record type }
{ also stores the length of the record not including this header }
{ All records start with this standard header except BOF and EOF records }
begin
  WriteWord(recid);
  WriteWord(reclen);
end;

procedure TExcelSheet.writeBOF;
{ marks the beginning of (sub)stream. Does not use a standard RecordHeader }
const
  vers  = $0006;             { version (unknown use, but its not XL version) }
  dt = $0000;                { substream type: $0010 = Worksheet }
  rupBuild = $0010;          { XL build no: unimportant(?) }
  rupYear = $096c;           { XL build year: unimportant(?) }
begin
  WriteByte(xlBOF);
  WriteByte(xlExcelVersion);
  WriteWord(vers);
  WriteWord(dt);
  WriteWord(rupBuild);
  WriteWord(rupYear);
end;

procedure TExcelSheet.writeEOF;
{ marks the end of a (sub)stream. Does not use a standard RecordHeader }
begin
  { EOF consists of the bytes $0a $00 $00 $00 }
  { we should not use WriteLongWord, as the bytes should not be re-arranged }
  WriteByte(xlEOF);
  WriteByte($00);
  WriteByte($00);
  WriteByte($00);
end;

procedure TExcelSheet.writeHeader;
{ stores the header to use on printouts }
var
  lnHeader :word;
begin
  lnHeader := length(Header)+1;
  WriteRecHeader(xlHeader, lnHeader);
  WriteText(Header, False);
end;

procedure TExcelSheet.writeFooter;
{ stores the footer to use on printouts }
var
  lnFooter :word;
begin
  lnFooter := length(Footer)+1;
  WriteRecHeader(xlFooter, lnFooter); { write recid and recsize }
  Writetext(Footer, False);
end;

procedure TExcelSheet.WriteLabel(aLabel :TLabelCell);
{ writes a LabelCell out to the file stream }
var
  lnLabel :word;
begin
  lnLabel := 8 + Length(aLabel.content);
  WriteRecHeader(xlLabel, lnLabel);
  WriteWord(aLabel.row);                   { zero based row of cell }
  WriteWord(aLabel.col);                   { zero based col of cell }
  WriteWord(aLabel.XFIndex);               { index to xlXF record }
  WriteText(aLabel.content, True);
end;

procedure TExcelSheet.WriteIndex;
{ an INDEX record contains internal cell scopes, }
{ and offsets to each DBCELL record in the file  }
{ This record type may require use of CONTINUE records }
var
  lnIndex : word;
begin
  lnIndex :=  16;
  WriteRecHeader(xlIndex, lnIndex);

  WriteLongWord(xlRESERVED);               { Reserved. must be 0 }
  WriteWord(Index.FirstRefdRow);         { first row which is referenced }
  WriteWord(Index.LastRow);              { last row + 1 which is defined in the sheet }
  WriteLongWord(xlRESERVED);               { Reserved. must be 0 }

  WriteWord($00);
  WriteWord($00);
end;

procedure TExcelSheet.WritePrintGridLines;
{ Writes a PRINTGRIDLINES record : applies to Excel sheet printouts }
const
  lnPrintGridLines = $02;
begin
  WriteRecHeader(xlPrintGridLines, lnPrintGridLines);
  WriteWord(Ord(PrintGridLines))
end;

procedure TExcelSheet.WriteXF(xfrec :TXF);
{ Writes XF record. Used to store extended formatting to rows and cells }
const
  lnXF = 12;
begin
  WriteRecHeader(xlXF, lnXF);

{  WriteWord(xfrec.XFFont); }
  WriteByte(xfrec.XFFormat);
  WriteByte(xfrec.XFInfo);
  WriteWord(xfrec.XFTextInfo);
  WriteWord(xfrec.FillColors);
  WriteWord(xfrec.BorderInfo);
  WriteWord(xfrec.BorderInfo2);
  WriteWord(xfrec.FillColors2);
end;

procedure TExcelSheet.WriteCodePage;
{ writes the environment codepage record }
const
  lnCodePage = $02;
begin
  WriteRecHeader(xlCodePage, lnCodePage);
  WriteWord(CodePage);
end;

procedure TExcelSheet.WriteCalcCount;
{ writes the Iteration count (as set in XL options dialog) }
const
  lnCalcCount = $02;
begin
  WriteRecHeader(xlCalcCount, lnCalcCount);
  WriteWord($0);
end;

procedure TExcelSheet.WriteDelta;
{ writes the Iteration delta record }
const
{  lnDelta = $04; }
  lnDelta = $08;
begin
  WriteRecHeader(xlDelta, lnDelta);
  WriteLongWord(Delta);
  WriteWord($0);
  WriteWord($14);
end;

procedure TExcelSheet.WriteLeftMargin;
{ specifies the left margin size in inches }
const
  lnLeftMargin = $08;
begin
  WriteRecHeader(xlLeftMargin, lnLeftMargin);
  WriteIEEE($0);
end;

procedure TExcelSheet.WriteRightMargin;
{ specifies the right margin size in inches }
const
  lnRightMargin = $08;
begin
  WriteRecHeader(xlRightMargin, lnRightMargin);
  WriteIEEE($0);
end;

procedure TExcelSheet.WriteBottomMargin;
{ specifies the bottom margin size in inches }
const
  lnBottomMargin = $08;
begin
  WriteRecHeader(xlBottomMargin, lnBottomMargin);
  WriteIEEE(BottomMargin);
end;

procedure TExcelSheet.WriteTopMargin;
{ specifies the top margin size in inches }
const
  lnTopMargin = $08;
begin
  WriteRecHeader(xlTopMargin, lnTopMargin);
  WriteIEEE(TopMargin);
end;

procedure TExcelSheet.WriteHCenter;
{ specifies if printouts should be centered horizontally }
const
  lnHCentering = $02;
begin
  WriteRecHeader(xlHCentering, lnHCentering);
  WriteWord(Ord(HCenter))
end;

procedure TExcelSheet.WriteVCenter;
{ specifies if printouts should be centered vertically }
const
  lnVCentering = $02;
begin
  WriteRecHeader(xlVCentering, lnVCentering);
  WriteWord(Ord(VCenter))
end;

procedure TExcelSheet.WriteIteration;
{ specifies if iteration option is enabled }
const
  lnIteration = $02;
begin
  WriteRecHeader(xlIteration, lnIteration);
  WriteWord(Ord(Iteration));
end;

procedure TExcelSheet.WritePrintHeaders;
{ specifies if sheet row & col headers are displayed }
const
  lnPrintHeaders = $04; { $02 }
begin
  WriteRecHeader(xlPrintHeaders, lnPrintHeaders);
  WriteWord($00);       { $00 }
  WriteWord(Ord(PrintHeaders));
end;

procedure TExcelSheet.WriteRow(aRow :TRow);
{ defines a row in the sheet }
const
  lnRow = 16;
begin
  WriteRecHeader(xlRow, lnRow);
  WriteWord(aRow.RowNo);
  WriteWord(aRow.FirstCol);
  WriteWord(aRow.LastCol1);
  WriteWord(aRow.RowHeight);
  WriteWord($0);                { used by excel to optimise loading file }
  WriteWord(xlRESERVED);

  if (aRow.XFIndex>0) then
    aRow.Attributes := aRow.Attributes or $80; { if XFindex applies then set fGhostDirty }

  WriteWord(aRow.Attributes);
  WriteWord(aRow.XFIndex)
end;

procedure TExcelSheet.WriteStyle(aStyle :TStyle);
{ outputs a given style record }
var
  lnStyle :word;
  tempWord :word;
begin
  if aStyle.IsBuiltIn then
    lnStyle := 4
  else
    lnStyle := 3 + Length(aStyle.Name);

  WriteRecHeader(xlStyle, lnStyle);

  tempword := aStyle.XFIndex;

{ set highest bit of word to flag whether style is Built In }
  if aStyle.IsBuiltIn then
    tempword := (tempword or $8000)
  else
    tempword := tempword and not $8000;

  WriteWord(tempword);

  if aStyle.IsBuiltIn then
  begin
    WriteByte(aStyle.BuiltIn);
    WriteByte(aStyle.Level);
  end
  else
  begin
    WriteText(aStyle.Name, False);
  end;
end;

procedure TExcelSheet.WriteDimensions;
const
  lnDimensions = 10;
begin
  WriteRecHeader(xlDimensions, lnDimensions);
  WriteWord(Dimensions.FirstRow); { first row which is referenced by a cell }
  WriteWord(Dimensions.LastRow);  { last row containing defined cells PLUS 1 }
  WriteWord(Dimensions.FirstCol); { first column containing a referenced cell }
  WriteWord(Dimensions.LastCol);  { last col containing defined cells PLUS 1 }
  WriteWord(xlRESERVED);          { RESERVED : must be zero }
end;

procedure TExcelSheet.WriteSetup;
{ writes a SETUP record. Used to specify options and }
{ measurements for the Page Setup }
const
  lnSetup = 34;
begin
  WriteRecHeader(xlSetup, lnSetup);

  with PageSetup do
  begin
    WriteWord(PaperSize);
    WriteWord(Scale);
    WriteWord(PageStart);
    WriteWord(FitWidth);
    WriteWord(FitHeight);

    WriteWord(Attributes);
    WriteWord(Resolution);
    WriteWord(VResolution);

{   WriteIEEE(HdrMargin);
    WriteIEEE(FtrMargin);   }

    { header margin }
    WriteWord($00);
    WriteWord($00);
    WriteWord($00);
    WriteByte($e0);
    WriteByte($3f);

    { footer margin }
    WriteWord($00);
    WriteWord($00);
    WriteWord($00);
    WriteByte($e0);
    WriteByte($3f);

    WriteWord(Copies);
  end;
end;

procedure TExcelSheet.WriteFNGroupCount;
{ Write an FNGROUPCOUNT record }
const
  lnFNGroupCount = $02;
begin
  WriteRecHeader(xlFNGroupCount, lnFNGroupCount);
  WriteWord($000e);
end;

procedure TExcelSheet.WriteWriteAccess;
{ Write a WRITEACCESS record. Stores the username of the person with write access }
const
  lnWriteAccess = 32;
begin
  WriteRecHeader(xlWriteAccess, lnWriteAccess);
  while Length(WriteAccess)<30 do       { Pad the string with spaces, to 31 chars }
    WriteAccess := WriteAccess + ' ';
  WriteText(WriteAccess, True);         { Write string, using a Word to specify the string length }
end;

procedure TExcelSheet.WriteGridSet;
{ Writes a GridSet record. Used to specify if the user has ever }
{ changed the gridlines state, in the Excel Page Setup (sheet tab) }
const
  lngridSet = $02;
begin
  WriteRecHeader(xlGridSet, lnGridSet);
  WriteWord($0000);
end;

procedure TExcelSheet.WriteDefRowHeight;
{ Writes a DEFROWHEIGHT record. Applies to all undefined rows }
const
  lnDefRowHeight = $02;
begin
  WriteRecHeader(xlDefRowHeight, lnDefRowHeight);
  WriteWord($0108);
end;

procedure TExcelSheet.WriteFormat(aFormat :TFormat);
{ Writes a FORMAT record }
var
  lnFormat :Word;
begin
  lnFormat := 3 + Length(aFormat.FormatStr);
  WriteRecHeader(xlFormat, lnFormat);

  WriteWord(aFormat.iFormat);
  WriteText(aFormat.FormatStr, False);  { Actual format string, uses 1 byte to specify string length }
end;

procedure TExcelSheet.WriteWindow1;
{ Writes a WINDOW1 record. Used to store information about Workbook-level window }
const
  lnWindow1 = 18;
begin
  WriteRecHeader(xlWindow1, lnWindow1);
  WriteWord(BookWindow.XPos);                 { Horizontal position of the window (in 1/20ths pt) }
  WriteWord(BookWindow.YPos);                 { Vertical position of the window (in 1/20ths pt) }
  WriteWord(BookWindow.Width);                { Width of the window (in 1/20ths pt) }
  WriteWord(BookWindow.Height);               { Height of the window (in 1/20ths pt) }
  WriteWord(BookWindow.Attributes);           { Windows flags: Hidden, Iconic, Visibility flags }
  WriteWord(BookWindow.CurrentTab);           { Zero-based index of the currently selected sheet tab }
  WriteWord(BookWindow.FirstTab);             { Zero-based index of the first visible sheet tab }
  WriteWord(BookWindow.NoOfSelectedTabs);     { Number of workbook sheet tabs currently selected }
  WriteWord(BookWindow.TabRatio);             { Ratio of widths of the workbook tabs to the horizontal scrollbar }
end;

procedure TExcelSheet.WriteWindow2;
{ Writes a WINDOW2 record. Used to store information about a WorkSheet-level window }
const
  lnWindow2 = 10;
begin
  WriteRecHeader(xlWindow2, lnWindow2);
  WriteWord(SheetWindow.Attributes);    { Flags:     Hidden, Iconic, Visibility flags }
  WriteWord(SheetWindow.TopRow);        { TopRow     Top-most row visible in the window }
  WriteWord(SheetWindow.LeftCol);       { LeftCol    Left-most column visible in window }
  WriteLongWord(SheetWindow.GridColor); { Colors     Row/column heading & Grid lines color. }
end;

procedure TExcelSheet.WriteFont(NewFont :TFontRec);
{ Writes a registered FONT record }
var
  lnFont :word;
begin
  lnFont := 7{15} + Length(NewFont.name);
  WriteRecHeader(xlFont, lnFont);

  WriteWord(NewFont.Height);           { Height of font (in 1/20ths pt) }
  WriteWord(NewFont.Attributes);       { Attributes.     Such as italic }
  WriteWord(NewFont.Color);            { Color.          Index to color palette for fonts color }
{  WriteWord(NewFont.Weight);           { BoldStyle.      $0190=Normal, $02bc=Bold, 100..1000=UserDefined }
{  WriteWord(NewFont.Script);           { Sub/SuperScript 0=None, 1=Sub, 2=Super }
{  WriteByte(NewFont.Underline);        { Underline.      0=None, 1=Sngl, 2=Dbl, $21=Sngl Acc, $22=Dbl Acc }
{  WriteByte(NewFont.Family);           { Font family.    From API TLogFont }
{  WriteByte(NewFont.CharSet);          { Font charset.   From API TLogFont }
{  WriteByte(xlRESERVED);               { RESERVED        Must be Zero }
  WriteText(NewFont.Name, False);      { Fonts name.     EG. 'Times New Roman' }
end;

procedure TExcelSheet.WriteCell(aCell :TCell);
{ given an instance of a LabelCell or NumberCell }
{ will call the correct cell types write method  }
begin
  case aCell.CellType of
    ctLabel  : WriteLabel(TLabelCell(aCell));
    ctNumber : WriteNumber(TNumberCell(aCell));
  end;
end;

procedure TExcelSheet.WriteToFile;
var
  loop :word;
  CurrentRow :integer;
  CurrentCell :longint;
{  StartRow, StartRowIndex,
  EndRow, EndRowIndex :integer; }
begin
  outFile := TFileStream.create(FileName, fmCreate);

  try
    RegisterMinimumFontSet;
    RegisterMinimumFormatSet;
    RegisterMinimumXFSet;
    RegisterMinimumStyleSet;

    WriteBOF;
    WriteWriteAccess;
    WriteIndex;
    WriteCodePage;
    WriteCalcMode;
    WriteCalcCount;
    WriteRefMode;
    WriteIteration;
    WriteDelta;
    WriteSaveRecalc;
{      WritePrecision; }
{      Write1904; }
    WritePrintHeaders;
    WritePrintGridLines;
    WriteGridSet;
    WriteGuts;
    WriteDefRowHeight;
    WriteCountry;

  { WriteHideObj; }
  { WriteWorkspaceBools; }

    { If the FontStore is not empty }
    if FontStore.count>0 then
    begin
     { write each Font in the FontStore }
      for loop := 0 to FontStore.count-1 do
        WriteFont(TFontRec(FontStore.At(loop)));
    end;

    WriteHeader;
    WriteFooter;

    WriteHCenter;
    WriteVCenter;
    WriteLeftMargin;
    WriteRightMargin;
    WriteTopMargin;
    WriteBottomMargin;
    WriteSetup;
    WriteBackup;
    WriteBuiltInFormatCount;

{      WriteFNGroupCount; }

    { Check if we have any Formats registered for the sheet, from the FormatStore }
    if FormatStore.count>0 then
    begin
      { If so, Write each Format registered for the sheet, from the FormatStore }
      for loop := 0 to FormatStore.Count-1 do
        WriteFormat(TFormat(FormatStore.At(loop)));
    end;

    WriteName;
    WriteWinprot;

    { Check if we have any XFs registered for the sheet, from the XFStore }
    if XFStore.count > 0 then
    begin
      { If so, Write each XF registered for the sheet, from the XFStore }
      for loop := 0 to XFStore.Count-1 do
        WriteXF(TXF(XFStore.At(loop)));
    end;

    { Check if we have any Styles registered for the sheet, from the StyleStore }
    if StyleStore.count > 0 then
    begin
      { If so, Write each Format registered for the sheet, from the FormatStore }
      for loop := 0 to StyleStore.Count-1 do
        WriteStyle(TStyle(StyleStore.At(loop)));
    end;

    { WritePalette; }
    WriteDefColWidth;

    { write out all of the column formatting objects }
    if ColInfoStore.Count > 0 then
      for loop := 0 to ColInfoStore.Count - 1 do
        WriteColInfo(TColInfo(ColInfoStore.Items[loop]));

    WriteDimensions;

    { write out all of the row objects }
    for CurrentRow := 0 to RowStore.count -1 do
      WriteRow(TRow(RowStore.At(CurrentRow)));

    { output the current CELL to the file }
    for CurrentCell := 0 to CellStore.count - 1 do
      WriteCell(TCell(CellStore.At(CurrentCell)));

    WriteWindow1;
    WriteWindow2;
    { WriteSCl; }
    WriteSelection;
    WriteRGCW;
    { WriteGCWCols; }
    { WriteDefColWidth; }
    WriteEOF;
  finally
    outFile.Free;
  end;
end;

function TFormatStore.RegisterFormat(aFormat :TFormat) :integer;
var
  NewFormat :TFormat;
  formatIndex :integer;
begin
 { Create a new format record. Copy values from given instance }
  NewFormat := TFormat.create;
  NewFormat.FormatStr := aFormat.FormatStr;
  NewFormat.iFormat := aFormat.iFormat;

  if (Duplicates = dupIgnore) then  { if restricting duplicates }
  begin
    formatindex := FindDuplicate(NewFormat);  { see if one exists }
    if (formatindex=-1) then
      formatindex := Add(NewFormat)           { if not, add new one }
  else
    NewFormat.Free;                       { otherwise destroy copy, and return existing index }
  end
  else
    formatindex := Add(NewFormat);          { if allowed duplicates, add regardless }

  RegisterFormat := formatindex;

end;

function TFontStore.RegisterFont(aFont :TFontRec) :Integer;
var
  NewFont :TFontRec;
  fontindex :integer;
begin
 { Create a new font record. Load in the values taken from the aFont parameter }
  NewFont := TFontRec.create;
  NewFont.Height := aFont.Height;
  NewFont.Attributes := aFont.Attributes;
  NewFont.Color := aFont.Color;
  NewFont.Weight := aFont.Weight;
  NewFont.Script := aFont.Script;
  NewFont.Underline := aFont.Underline;
  NewFont.Family := aFont.Family;
  NewFont.Charset := aFont.Charset;
  NewFont.Name := aFont.Name;

  if (Duplicates = dupIgnore) then              { if restricting duplicates }
  begin
    fontindex := FindDuplicate(NewFont);  { see if one exists }
    if (fontindex=-1) then
      fontindex := Add(NewFont)           { if not, add new one }
  else
    NewFont.Free;                         { otherwise destroy copy, and return existing index }
  end
  else
    fontindex := Add(NewFont);            { if allowed duplicates, add regardless }

  RegisterFont := fontindex;
end;

function TXFStore.RegisterXF(anXF :TXF) :Integer;
var
  NewXF :TXF;
  XFindex :integer;
begin
 { Create a new XF record. Load in the values taken from the anXF parameter }
  NewXF := TXF.create;

{  NewXF.XFFont := anXF.XFFont; }
  NewXF.XFFormat := anXF.XFFormat;
  NewXF.XFInfo := anXF.XFInfo;
  NewXF.XFTextInfo := anXF.XFTextInfo;
  NewXF.FillColors := anXF.FillColors;
  NewXF.BorderInfo := anXF.BorderInfo;
  NewXF.BorderInfo2 := anXF.BorderInfo2;
  NewXF.FillColors2 := anXF.FillColors2;

  if (Duplicates = dupIgnore) then              { if restricting duplicates }
  begin
    XFindex := FindDuplicate(NewXF);
    if (XFindex=-1) then
      XFindex := Add(NewXF)
    else
      NewXF.Free
  end
  else
    XFindex := Add(NewXF);

  RegisterXF := XFindex;
end;

procedure TExcelSheet.WriteNumber(aNumberCell :TNumberCell);
const
  lnNumber = 14;
begin
  WriteRecHeader(xlNumber, lnNumber);
  WriteWord(aNumberCell.Row);
  WriteWord(aNumberCell.Col);
  WriteWord(aNumberCell.XFIndex);                 { XF INDEX }
  WriteIEEE(aNumberCell.Content);
end;

procedure TExcelSheet.WriteDBCell(aDBCell :TDBCell);
var
  lnDBCell :integer;
  iOffset :integer;
begin
  { length of record = 4 bytes for Block Offset + 2 bytes for each Cell Offset }
  lnDBCell := 4 + (2 * aDBCell.FCellOffsets.count);

  WriteRecHeader(xlDBCell, lnDBCell);
  WriteLongWord(aDBCell.FBlockOffset);

  for iOffset := 0 to aDBCell.FCellOffsets.count - 1 do
    WriteWord(TOffset(aDBCell.FCellOffsets.At(iOffset)).FOffset);
end;

{ ############################## }
{     MISCELLANEOUS ROUTINES     }
{ ############################## }

procedure TExcelSheet.SetBit(var InByte, iBit :byte; State: Boolean);
{ sets a bit (0-7) of a byte to State }
begin
  if (iBit>=0) and (iBit<=7) then                      { if the bit index is valid }
    InByte := (InByte and (not iBit)) or (1 shl iBit); { clears the bit then loads it with State }
end;

function TFontStore.FindDuplicate(ThisFont :TFontRec) :Integer;
{ Searches for a duplicate of ThisFont in the FontStore. }
{ Returns the duplicate Fonts's index if found, else -1. }
var
  EachFont :integer;
  Duplicate :Boolean;
  aFont :TFontRec;
begin
  EachFont := 0;
  Duplicate := False;

  { if we actually already have some fonts registered in the store }
  if (count>0) then
  begin
    { then loop over each one }
    for EachFont := 0 to Count-1 do
    begin
      { get current font from the store }
      aFont := TFontRec(At(EachFont));

      { assume we have found a duplicate until we find a field which differs }
      Duplicate := True;

      if ThisFont.Height <> aFont.Height         then Duplicate := False;
      if ThisFont.Attributes <> aFont.Attributes then Duplicate := False;
      if ThisFont.Color <> aFont.Color           then Duplicate := False;
      if ThisFont.Weight <> aFont.Weight         then Duplicate := False;
      if ThisFont.Underline <> aFont.Underline   then Duplicate := False;
      if ThisFont.Family <> aFont.Family         then Duplicate := False;
      if ThisFont.CharSet <> aFont.CharSet       then Duplicate := False;
      if ThisFont.Name <> aFont.Name             then Duplicate := False;

      { if after testing all fields, we have a duplicate then stop searching }
      if Duplicate then break;
    end;
  end;

  { if we found an exact copy then return its index }
  if Duplicate then
    FindDuplicate := EachFont
  { otherwise return -1 signifying that no duplicate was found }
  else
    FindDuplicate := -1;
end;

procedure TExcelSheet.RegisterMinimumFontSet;
var
  aFont :TFontRec;
  RegLoop :Integer;
begin
  aFont := TFontRec.create;

 { set values common to all 5 instances }
  aFont.Height := $00c8;
  aFont.Color := $7fff;
  aFont.Weight := $0190;
  aFont.Script := $0000;
  aFont.UnderLine := $00;
  aFont.Family := $00;
  aFont.CharSet := $a5;
  aFont.Name := 'Arial';

  FontStore.Duplicates := dupAccept;    { minimum set may contain duplicates }

 { write the 5 minimum font records (Arial) }
  for RegLoop := 0 to 4 do
  begin
    if (RegLoop=4) then
      aFont.Attributes := 0
    else
      aFont.Attributes := RegLoop;

    FontStore.RegisterFont(aFont);

    if (aFont.Weight=$0190) then
      aFont.Weight := $02bc
    else
      aFont.Weight := $0190;
  end;

  aFont.Free;
  FontStore.Duplicates := dupAccept;
end;

procedure TExcelSheet.RegisterMinimumStyleSet;
var
  aStyle :TStyle;
begin
  aStyle := TStyle.create;
  StyleStore.Duplicates := dupAccept;

  aStyle.XFIndex := $10;
  aStyle.IsBuiltIn := True;
  aStyle.BuiltIn := $03;
  aStyle.Level := $FF;
  aStyle.Name := '';
  StyleStore.RegisterStyle(aStyle);

  aStyle.XFIndex := $11;
  aStyle.BuiltIn := $06;
  StyleStore.RegisterStyle(aStyle);

  aStyle.XFIndex := $12;
  aStyle.BuiltIn := $04;
  StyleStore.RegisterStyle(aStyle);

  aStyle.XFIndex := $13;
  aStyle.BuiltIn := $07;
  StyleStore.RegisterStyle(aStyle);

  aStyle.XFIndex := $00;
  aStyle.BuiltIn := $00;
  StyleStore.RegisterStyle(aStyle);

  aStyle.XFIndex := $14;
  aStyle.BuiltIn := $05;
  StyleStore.RegisterStyle(aStyle);

  StyleStore.Duplicates := dupAccept;
  aStyle.Free;
end;

procedure TExcelSheet.RegisterMinimumFormatSet;
var
  aFormat :TFormat;
begin
  aFormat := TFormat.Create;
  FormatStore.Duplicates := dupAccept;

  { General format - used for any cell }
  { if a cell is published as a string cell, but actually contains a numeric value }
  { ie '2.22' }
  { and the cell uses General format, then once in Excel, the numeric can be }
  { evaluated as if the cell was published as a Number cell, once in Excel }
  aFormat.FormatStr := 'General';
  aFormat.iFormat := $0007;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := '0.00';
  aFormat.iFormat := $0004;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := '"#"#,##0;\-"#"#,##0';
  aFormat.iFormat := $0005;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := '"#"#,##0;[Red]\-"#"#,##0';
  aFormat.iFormat := $0006;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := '"#"#,##0.00;\-"#"#,##0.00';
  aFormat.iFormat := $0007;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := '"#"#,##0.00;[Red]\-"#"#,##0.00';
  aFormat.iFormat := $0008;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := ' -"#"* #,##0 -;\-"#"* #,##0 -; -"#"* "-" -; -@ -';
  aFormat.iFormat := $002a;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := ' -* #,##0 -;\-* #,##0 -; -* "-" -; -@ -';
  aFormat.iFormat := $0029;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := ' -"#"* #,##0.00 -;\-"#"* #,##0.00 -; -"#"* "-"?? -; -@ -';
  aFormat.iFormat := $002c;
  FormatStore.RegisterFormat(aFormat);

  aFormat.FormatStr := ' -* #,##0.00 -;\-* #,##0.00 -; -* "-"?? -; -@ -';
  aFormat.iFormat := $002b;
  FormatStore.RegisterFormat(aFormat);

  FormatStore.Duplicates := dupIgnore;
  aFormat.Free;
end;

procedure TExcelSheet.RegisterMinimumXFSet;
var
  anXF :TXF;
  loop :Integer;
begin
  { create a temporary XF }
  anXF := TXF.Create;

  { accept duplicates (for the minimum set only) } 
  XFStore.Duplicates := dupAccept;

  { For the bare minimum of 21 XFs }
  { plus the Bold XF }
  for loop := 1 to 22 do
  begin
    case loop of
      1: begin
           anXF.XFFormat := $0;
           anXF.XFInfo := $0000;
           anXF.XFTextInfo := $fff5;
           anXF.FillColors := $0120;
           anXF.BorderInfo := $ce00;
           anXF.BorderInfo2 := $0000;
         end;
      2: begin
           anXF.XFFormat := $0001;
           anXF.FillColors := $f520;
         end;
      4: anXF.XFFormat := 002;
      6: anXF.XFFormat := $0000;
      14: anXF.FillColors := $f420;

      16: begin
            anXF.XFTextInfo := $0001;
            anXF.FillColors := $0020;
          end;
      17: begin
            anXF.XFFormat := $0005;
            anXF.XFInfo := $21;            {002b} {used by Excel Worksheet v5.00}
            anXF.XFTextInfo := $fff5;
            anXF.FillColors := $f820;
          end;
      18: anXF.XFInfo := $1f;              {0029} {used by Excel Worksheet v5.00}
      19: anXF.XFInfo := $20;              {002c} {used by Excel Worksheet v5.00}
      20: anXF.XFInfo := $1e;              {002a} {used by Excel Worksheet v5.00}
      21: anXF.XFInfo := $0d;              {0009} {used by Excel Worksheet v5.00}

      {W8 START -- Extended Format record for TextStyle = tsBOLD } 
      22: begin
            anXF.XFFormat := $01;
            anXF.XFInfo := $00;
            anXF.XFTextInfo := $0001;
          end;
      {W8 END}
    end;

    { add the XF to the XF store }
    XFStore.RegisterXF(anXF);
  end;

  { now all minimum XF's are in the store, take no more duplicates on board }
  XFStore.Duplicates := dupIgnore;

  { destroy the tempoary XF }
  anXF.Free;
end;

procedure TExcelSheet.AddCell(NewCell :TCell);
var
  iRow :Integer;
  CellsRow :TRow;
  RowLoop :Integer;
  FoundRow :Boolean;
begin
  if ((NewCell.Row < FMaxRows) and (NewCell.Col < FMaxCols)) then
  begin
    inc(FNoOfCells);

    { Update the extreme cell boundaries in DIMENSIONS }
    if (NewCell.Row < Dimensions.FirstRow) then Dimensions.FirstRow := NewCell.Row;
    if ((NewCell.Row+1) > Dimensions.LastRow)  then Dimensions.LastRow := NewCell.Row+1;
    if (NewCell.Col < Dimensions.FirstCol) then Dimensions.FirstCol := NewCell.Col;
    if ((NewCell.Col+1) > Dimensions.LastCol)  then Dimensions.LastCol := NewCell.Col+1;

    { Update the extreme cell boundaries in INDEX }
    if (NewCell.Row < Index.FirstRefdRow) then Index.FirstRefdRow := NewCell.Row;
    if ((NewCell.Row+1) > Index.LastRow) then Index.LastRow := NewCell.Row+1;

    { try to find an existing ROW record for this Row }
    { if one does not exist then create and register it }
    FoundRow := False;
    RowLoop := 0;
    iRow := 0;

    if RowStore.count > 0 then
    begin
      for RowLoop := 0 to RowStore.count - 1 do
      begin
        if TRow(RowStore.At(RowLoop)).RowNo = NewCell.Row then
        begin
          FoundRow := True;
          iRow := RowLoop;
          break;
        end;
      end;
    end;

    if (FoundRow) then
    begin
    { check if columns are more extreme than those currently in row }
    { if they are, then modify row object to record new extremes }
      if NewCell.Col < TRow(RowStore.At(RowLoop)).FirstCol then
        TRow(RowStore.At(RowLoop)).FirstCol := NewCell.Col;

      if (NewCell.Col+1) > TRow(RowStore.At(RowLoop)).LastCol1 then
        TRow(RowStore.At(RowLoop)).LastCol1 := (NewCell.Col+1);
    end
    else
    begin
    { add new row with column extremes taken from cell }
      inc(FNoOfRows);
      CellsRow := TRow.create;
      CellsRow.RowNo := NewCell.Row;
      CellsRow.XFIndex := NewCell.XFIndex;
      CellsRow.FirstCol := NewCell.Col;
      CellsRow.LastCol1 := NewCell.Col+1;
      iRow := RowStore.RegisterRow(CellsRow);

      CellsRow.Free;
    end;

    { iRow now contains index of new (or already existing) row record }
    { check ROWs column scopes and modify if new cell exceeds them }
     if (NewCell.Col < TRow(RowStore.At(iRow)).FirstCol) then
        TRow(RowStore.At(iRow)).FirstCol := NewCell.Col;
     if ((NewCell.Col+1) > TRow(RowStore.At(iRow)).LastCol1) then
        TRow(RowStore.At(iRow)).LastCol1 := NewCell.Col+1;

    { ** If (no of rows > 32) then need to create new DBCELL record }
    {    we then need to update index record with the new DBCELL's offset }
    { ** The above comment is applicable only for Excel v5.00 worksheet and }
    {    not currently implemented }

    { add the actual cell object to the CELLSTORE }
    CellStore.Add(NewCell);  
  end;
end;

function TXFStore.FindDuplicate(ThisXF :TXF) :Integer;
{ Searches for a duplicate of ThisXF in the XFStore. }
{ Returns XF's index if found, else -1. }
var
  EachXF :integer;
  Duplicate :Boolean;
begin
  Duplicate := False;
  EachXF := -1;

  { if there are no XF's in the XFStore, skip over and return -1 }
  if (count > 0) then
  begin
    { examine each XF in the XFStore }
    for EachXF := 0 to Count-1 do
    begin
     { assume the current XF matches thisXF }
      Duplicate := True;
      { until one of the fields differs }
      if ThisXF.XFFormat    <> TXF(At(EachXF)).XFFormat    then Duplicate := False;
      if ThisXF.XFInfo      <> TXF(At(EachXF)).XFInfo      then Duplicate := False;
      if ThisXF.XFTextInfo  <> TXF(At(EachXF)).XFTextInfo  then Duplicate := False;
      if ThisXF.FillColors  <> TXF(At(EachXF)).FillColors  then Duplicate := False;
      if ThisXF.BorderInfo  <> TXF(At(EachXF)).BorderInfo  then Duplicate := False;
      if ThisXF.BorderInfo2 <> TXF(At(EachXF)).BorderInfo2 then Duplicate := False;
      if ThisXF.FillColors2 <> TXF(At(EachXF)).FillColors2 then Duplicate := False;
      { if we have found a duplicate, stop searching XFStore, and return index }
      if Duplicate then break;
    end;
  end;
  if Duplicate then
    FindDuplicate := EachXF               { a match was found. Return index of duplicate }
  else
    FindDuplicate := -1;                  { no match exists in the XFStore, return -1 }
end;

function TRowStore.RegisterRow(aRow :TRow) :integer;
var
  iRow :Integer;
  NewRow :TRow;
  Duplicate :Boolean;
{  str :string; }
begin
  Duplicate := False;
  Result := -1;

  { see if row with this rowno exists }
  if (count>0) then
  begin
    for iRow := 0 to count-1 do
    begin
      if (TRow(At(iRow)).RowNo = aRow.RowNo) then
      begin
        RegisterRow := iRow;
        Duplicate := True;
        break;
      end;
    end
  end;

  if (not Duplicate) then
  begin
  { take a copy of the given row }
    NewRow := TRow.Create;
    NewRow.RowNo := aRow.RowNo;
    NewRow.FirstCol := aRow.FirstCol;
    NewRow.LastCol1 := aRow.LastCol1;
    NewRow.RowHeight := aRow.RowHeight;
    NewRow.Attributes := aRow.Attributes;
    NewRow.XFIndex := aRow.XFIndex;

    Result := Add(NewRow);
  end;
end;

function TStyleStore.RegisterStyle(aStyle :TStyle) :integer;
var
  NewStyle :TStyle;
  StyleIndex :Integer;
begin
  StyleIndex := FindDuplicate(aStyle);

  if (StyleIndex = -1) then
  begin
    NewStyle := TStyle.Create;
    NewStyle.XFIndex := aStyle.XFIndex;
    NewStyle.IsBuiltIn := aStyle.IsBuiltIn;
    NewStyle.BuiltIn := aStyle.BuiltIn;
    NewStyle.Level := aStyle.Level;
    NewStyle.Name := aStyle.Name;

    StyleIndex := Add(NewStyle)
  end;
{  else
    NewStyle.Free; }

  RegisterStyle := StyleIndex;
end;

function TStyleStore.FindDuplicate(ThisStyle :TStyle) :Integer;
{ Searches for a duplicate of ThisStyle in the StyleStore. }
{ Returns Style's index if found, else -1. }
var
  EachStyle :integer;
  Duplicate :Boolean;
begin
  EachStyle := 0;
  Duplicate := False;

  { if there are no Style's in the StyleStore, skip over and return -1 }
  if (count>0) then
  begin
    { examine each Style in the StyleStore }
    for EachStyle := 0 to Count-1 do
    begin
     { assume the current Style matches thisStyle }
      Duplicate := True;

      { until one of the fields differs }
      if ThisStyle.XFindex <> TStyle(At(EachStyle)).XFindex then Duplicate := False;
      if ThisStyle.BuiltIn <> TStyle(At(EachStyle)).BuiltIn then Duplicate := False;
      if ThisStyle.Level   <> TStyle(At(EachStyle)).Level   then Duplicate := False;
      if ThisStyle.Name    <> TStyle(At(EachStyle)).Name    then Duplicate := False;

      { if we have found a duplicate, stop searching StyleStore, and return index }
      if Duplicate then break;
    end;
  end;
  if Duplicate then
    FindDuplicate := EachStyle            { a match was found. Return index of duplicate }
  else
    FindDuplicate := -1;                  { no match exists in the StyleStore, return -1 }
end;

function TFormatStore.FindDuplicate(ThisFormat :TFormat) :Integer;
{ Searches for a duplicate of ThisFormat in the FormatStore. }
{ Returns existing format's index if found, else -1. }
var
  EachFormat :integer;
  Duplicate :Boolean;
begin
  EachFormat := 0;
  Duplicate := False;

  { if there are no Format's in the StyleStore, skip over and return -1 }
  if (count>0) then
  begin
    { examine each Format in the StyleStore }
    for EachFormat := 0 to Count-1 do
    begin
     { assume the current Format matches thisFormat }
      Duplicate := True;

      { until one of the fields differs }
      if ThisFormat.FormatStr <> TFormat(At(EachFormat)).FormatStr then Duplicate := False;
      if ThisFormat.iFormat <> TFormat(At(EachFormat)).iFormat then Duplicate := False;

      { if we have found a duplicate, stop searching FormatStore, and return index }
      if Duplicate then break;
    end;
  end;
  if Duplicate then
    FindDuplicate := EachFormat           { a match was found. Return index of duplicate }
  else
    FindDuplicate := -1;                  { no match exists in the FormatStore, return -1 }
end;

destructor TExcelSheet.Destroy;
{ Frees up all the objects stored in the TExcelSheet }
begin
 { free up the DBCell }
  DBCell.Free;

 { free up the container classes, and contained object }
  FormatStore.Free;
  StyleStore.Free;
  FontStore.Free;
  ColInfoStore.Free;
  XFStore.Free;
  RowStore.Free;
  CellStore.Free;

  inherited Destroy;
end;

procedure TExcelSheet.AddText(iRow, iCol :Integer; str: string);
{ creates a LabelCell and adds it to the CELLSTORE }
var
  newCell :TLabelCell;
begin
  if (str <> '') then {W8}
  begin
    newCell := TLabelCell.Create;

    with newCell do
    begin
      Row := iRow;
      Col := iCol;

      if (FTextStyle = tsBold) then
        XFIndex := xlBOLD_XF
      else
        XFIndex := xlNORMAL_XF;

      CellType := ctLabel;
      Content := str;
    end;

{WBF131}{ CellStore.Add(newCell);}
{WBF131} AddCell(NewCell);
  end;
end;

procedure TExcelSheet.AddNumber(iRow, iCol :Integer; num: double);
{ creates a NumberCell and adds it to the CELLSTORE }
var
  newCell :TNumberCell;
begin
  newCell := TNumberCell.Create;

  with newCell do
  begin
    Row := iRow;
    Col := iCol;

    if (FTextStyle = tsBold) then
      XFIndex := xlBOLD_XF
    else
      XFIndex := xlNORMAL_XF;

    CellType := ctNumber;
    Content := num;
  end;

{WBF131}{ CellStore.Add(newCell);}
{WBF131} AddCell(NewCell);
end;

procedure TExcelsheet.WriteDefColWidth;
const
  lnStdColWidth = 2;
begin
  WriteRecHeader(xlDefColWidth, lnStdColWidth);
  WriteWord(FDefColWidth);
end;

procedure TExcelSheet.LoadFromFile(aStringGrid :TStringGrid; filename :string);
begin
  ReadFromFile(filename);
  FeedIntoGrid(aStringGrid);
end;

procedure TExcelSheet.ReadFromFile(filename :string);
begin
  try
    try
      { create a file stream to hold the excel sheet }
      InFile := TMemoryStream.create;

      { load the file into the memory stream }
      InFile.LoadFromFile(filename);

      InFile.Position := 0;

      { keep reading until we have reached the end of the file }
      while (InFile.position < InFile.size) do
        ReadRecHeader;
    except
      on E :Exception do
        showmessage('Error reading excel file: ' + filename);
     end;
  finally
    { destroy the memory stream }
    InFile.Free;
  end;
end;

procedure TExcelSheet.ReadRecHeader;
var
  recid, reclen :word;
begin
  recid  := ReadWord;

  if (lo(recid) = xlBOF) then  { see if xlBOF }
  begin
    ReadWord;
    ReadWord;
    ReadWord;
  end
  else
  if (recid = xlEOF) then  { see if xlEOF }
    ReadWord
  else
  begin
    reclen := ReadWord;    { get length for normal record }

    case recid of
      { see if the record requires some sort of processing }
      xlNumber: ReadNumber;
      xlLabel:  ReadLabel;

      xlFont:
      begin
{        Watch('Font'); }
        ReadAndDiscard(RecLen);
      end;

      xlWriteAccess:
      begin
{        Watch('WriteAccess'); }
        ReadAndDiscard(RecLen);
      end;
      { IEEE }
      { RK }
    else
      { if we are not going to do anything special with this }
      { record then just ignore it }
      ReadAndDiscard(reclen);
    end;
  end;
end;


procedure TExcelSheet.ReadNumber;
{ Reads a number record from the stream, and adds it to the cell store }
var
  Row, Col :word;
  Num :double;
begin
  Row := ReadWord;
  Col := ReadWord;
  Num := ReadIEEE;

  AddNumber(Row, Col, Num);
end;

procedure TExcelSheet.ReadLabel;
{ Reads a label record from the stream, and adds it to the cell store }
var
  Row, Col :word;
  tmpStr :string;
begin
  Row := ReadWord;
  Col := ReadWord;
  tmpStr := ReadText(True);

  addText(Row, Col, tmpStr);
end;

procedure TExcelSheet.ReadAndDiscard(noOfBytes :word);
var
  loop :word;
begin
{  watch('Discarded: ' + IntToStr(noOfBytes)); }
  for loop := 1 to noOfBytes do
    ReadByte;
end;

function TExcelSheet.ReadByte :Byte;
{ reads a byte from the stream. All file-read routines call this in turn }
var
  inbyte :byte;
begin
  infile.read(inbyte, 1);
  result := inbyte;
end;

function TExcelSheet.ReadWord :word;
{ reads in a word in little endian format }
var
  x :word;
  y :word;
begin
  x := ReadByte;
  y := x;
  x := ReadByte;
  y := y + (x*256);

  result := y;
end;

function TExcelSheet.ReadLongWord :longint;
{ reads in the longword in little endian format }
var
  y :longint;
  x :longint;
begin
  y := ReadWord;
  x := y;
  y := ReadWord;
  x := x + (y*65535);

  result := x;
end;

function TExcelSheet.ReadIEEE :double;
{ Reads in an IEEE 80x87 format floating point number to the stream }
var
  IEEE :TDoubleRec;
  loop :Integer;
begin
  for loop := 1 to 8 do
    IEEE.byteview[loop] := ReadByte;
  result := IEEE.doubleview;
end;

function TExcelSheet.ReadText(WordLen: Boolean) :string;
{ Reads a string from the stream, using a preceding word to store its length }
var
  byteno :byte;
  strlength :integer;
  returnstr :string;
begin
  if WordLen then
    strlength := ReadWord          { use word for length }
  else
    strlength := ReadByte;         { just use a byte }

  for byteno := 1 to strlength do    { write string contents }
    returnStr[byteno] := char(ReadByte);

  result := returnStr;
end;

procedure TExcelSheet.FeedIntoGrid(aGrid :TStringGrid);
{ Fills any given string grid with the cells currently in the ExcelSheet }
var
  Row, Col :word;
  loop :Integer;
  tmpint :Integer;
begin
  { set the dimensions of the StringGrid }
  tmpint := Dimensions.LastRow;
  dec(tmpint);

  if (tmpint > 0) then
    aGrid.ColCount := tmpint;

  tmpint := Dimensions.LastCol;
  dec(tmpint);

  if (tmpint > 0) then
    aGrid.RowCount := tmpint;

  { for each of the cell objects }
  for loop := 0 to CellStore.count - 1 do
  begin
    { retrieve the row and column which this cell represents }
    Row := TCell(CellStore.At(loop)).Row;
    Col := TCell(CellStore.At(loop)).Col;

    { load the cells objects content into the string grid }
    case TCell(CellStore.At(loop)).CellType of
      ctNumber:
      begin
        aGrid.Cells[Row, Col] :=
          GP_Real2Str(TNumberCell(CellStore.At(loop)).Content, 2);
      end;

      ctLabel:
      begin
        aGrid.Cells[Row, Col] :=
          TLabelCell(CellStore.At(loop)).Content;
      end;
    end;
  end;
end;

{ ## procedures to implement the use of a DBCELL class }
{ tracks a variety of stream offsets backwards to the start of 2nd row in each block }
{ along with a series of offsets back to the start of each cell for the current block }
procedure TDBCell.ResetOffsets;
var
  iOffset :integer;
begin
  for iOffset := 0 to FCellOffsets.count - 1 do
    TOffset(FCellOffsets).FOffset := 0;
end;

procedure TDBCell.SetOffset(cellno, offset :integer);
{ given an index into the cellstore, for a particular cell }
{ will set the file offset for this cell. Used to track a list of offsets back }
{ from the start of a DBCELL record, to the start of each cell in the current ROW }
begin
  if (cellno <= FCellOffsets.count-1) then
    TOffset(FCellOffsets.At(cellno)).FOffset := Offset;
end;

procedure TDBCell.BuildCellOffsetList(NoOfCells :integer);
var
  iCell :integer;
  newOffset :TOffset;
begin
  { for each cell in the cell list, add the a place holder for that cells offset }
  for iCell := 1 to NoOfCells do
  begin
    newOffset := TOffset.create;
    FCellOffsets.Add(newOffset);
  end;
end;

procedure TExcelSheet.WriteCalcMode;
{ writes the calculation mode record }
const
  lnCalcMode = $02;
begin
  WriteRecHeader(xlCalcMode, lnCalcMode);
  WriteWord(Ord(CalcMode));
end;

procedure TExcelSheet.WriteProtect;
{ writes the protection record }
const
  lnProtect = $02;
begin
  WriteRecHeader(xlProtect, lnProtect);
  WriteWord(Ord(Protect));
end;

procedure TExcelSheet.WriteWindowProtect;
{ writes the window protect record }
const
  lnWindowProtect = $02;
begin
  WriteRecHeader(xlWindowProtect, lnWindowProtect);
  WriteWord(Ord(WindowProtect));
end;

procedure TExcelSheet.WriteBackup;
{ writes the backup record }
const
  lnBackup = $02;
begin
  WriteRecHeader(xlBackup, lnBackup);
  WriteWord(Ord(Backup));
end;

procedure TExcelSheet.WriteSaveRecalc;
{ writes the recalc before save record }
const
  lnSaveRecalc = $02;
begin
  WriteRecHeader(xlSaveRecalc, lnSaveRecalc);
  WriteWord(Ord(SaveRecalc));
end;

procedure TExcelSheet.WriteGuts;
{ writes the gutter values record }
const
  lnGuts = $08;
begin
  WriteRecHeader(xlGuts, lnGuts);
  WriteWord(Guts);
  WriteWord(Guts);
  WriteWord(Guts);
  WriteWord(Guts);
end;

procedure TExcelSheet.WriteWSBool;
{ writes the WSBOOL record }
const
  lnWSBool = $02;
begin
  WriteRecHeader(xlWSBool, lnWSBool);
  WriteWord(WSBool);
end;

procedure TExcelSheet.WriteRefMode;
{ writes the Ref mode record }
const
  lnRefMode = $02;
begin
  WriteRecHeader(xlRefMode, lnRefMode);
  WriteWord(Ord(RefMode));
end;

procedure TExcelSheet.WriteCountry;
{ writes the country record }
const
  lnCountry = $04;
begin
  WriteRecHeader(xlCountry, lnCountry);
  WriteWord($01);
  WriteWord($2c);
end;

procedure TExcelSheet.WriteSelection;
{ writes the selection record }
const
  lnSelection = $0f;
begin
  WriteRecHeader(xlSelection, lnSelection);
  WriteByte($03);
  WriteByte($13);
  WriteByte($00);
  WriteByte($00);
  WriteByte($00);
  WriteByte($00);
  WriteByte($00);
  WriteByte($01);
  WriteByte($00);
  WriteByte($13);
  WriteByte($00);
  WriteByte($13);
  WriteByte($00);
  WriteByte($00);
  WriteByte($00);
end;

procedure TExcelSheet.WriteRGCW;
{ writes the selection record }
var
  padding :integer;
const
  lnRGCW = $22;
begin
  WriteRecHeader(xlRGCW, lnRGCW);
  WriteByte($20);
  WriteByte($00);
  for padding := 1 to 16 do
    WriteWord($ffff);
end;

procedure TExcelSheet.WriteName;
{ writes the name record }
const
  lnName = $07;
begin
  WriteRecHeader(xlName, lnName);
  WriteWord($20);
  WriteWord($100);
  WriteWord($00);
  WriteByte($04);
end;

procedure TExcelSheet.WriteWinProt;
{ writes the win protection record }
const
  lnWinprot = $02;
begin
  WriteRecHeader(xlWinProt, lnWinProt);
  WriteWord($00);
end;

procedure TExcelSheet.WriteBuiltInFormatCount;
const
  lnBuiltInFormatcount = $02;
begin
  WriteRecHeader(xlBuiltInFormatCount, lnBuiltInFormatCount);
  WriteWord($1b);
end;

procedure TCellStore.CellStoreOnCompare(ItemToInsert, Key : TObject;
   Var Index : TCompareResult; Var PartialResult : Boolean);
begin
  { sort cells inascending order by row, by column }
  if (TCell(ItemToInsert).Row < TCell(Key).Row) then
    Index := crLessThan
  else
  if (TCell(ItemToInsert).Row > TCell(Key).Row) then
    Index := crGreaterThan
  else
  if (TCell(ItemToInsert).Row = TCell(Key).Row) then
  begin
    if (TCell(ItemToInsert).Col < TCell(Key).Col) then
      Index := crLessThan
    else
    if (TCell(ItemToInsert).Col > TCell(Key).Col) then
      Index := crGreaterThan
    else
      { a cell is considered to be duplicate if row and col values are equal }
      Index := crEqual;
  end;
end;

procedure TRowStore.RowStoreOnCompare(ItemToInsert, Key : TObject;
   Var Index : TCompareResult; Var PartialResult : Boolean);
begin
  if (TRow(ItemToInsert).RowNo > TRow(Key).RowNo) then
    Index := crGreaterThan
  else
  if (TRow(ItemToInsert).RowNo < TRow(Key).RowNo) then
    Index := crLessThan
  else
    Index := crEqual;
end;

function TExcelSheet.ValidateSheetSize(rows, cols :longint) :TXLValidation;
{ given the maximum propsed size of the sheet, this will validate the sheet }
{ size, and return a validation result, enabling the object which created the }
{ excelsheet to determine if the sheet is a valid size }
begin
  if (rows > FMaxRows) then
  begin
    if (cols > FMaxCols) then
      result := xvSheetTooLarge
    else
      result := xvTooManyRows;
  end
  else
  if (cols > FMaxCols) then
    result := xvTooManyCols
  else
    result := xvOK;
end;

{W8 START}
constructor TColInfo.create;
begin
  inherited create;

  XFIndex := $000f;
  Options := $0000;
  ColWidth := 0;
end;

procedure TExcelSheet.WriteColInfo(ColInfo :TColInfo);
const
  lnColInfo = $000c;
begin
  WriteRecHeader(xlColInfo, lnColInfo);

  with ColInfo do
  begin
    WriteWord(Column);
    WriteWord(Column);
    WriteWord(ColWidth);
    WriteWord(XFIndex);
    WriteWord(Options);
    WriteWord(xlRESERVED);
  end;
end;

procedure TColInfoStore.AddColInfo(startcol, endcol, colwidth :word);
var
  tmpColInfo :TColInfo;
begin
  tmpColInfo := TColInfo.create;

  tmpColInfo.Column := startcol;
  tmpColInfo.ColWidth := ColWidth;
  Add(tmpColInfo);
end;

function TExcelSheet.getColumnWidth(index :integer) :word;
var
  tmpColInfo :TColInfo;
begin
  { return value of $FFFF means that column uses default column width }
  result := $FFFF;

  tmpColInfo := ColInfoStore.FindColInfo(index);
  if assigned(tmpColInfo) then
    result := tmpColInfo.ColWidth;
end;

procedure TExcelSheet.setColumnWidth(index :integer; ColumnWidth :word);
var
  tmpColInfo :TColInfo;
begin
  { see if colinfo is already created for this column }
  tmpColInfo := ColInfoStore.FindColInfo(index);

  { if we already have a colinfo object for this column }
  if assigned(tmpColInfo) then
    { then revise existing objects colwidth value }
    tmpColInfo.ColWidth := ColumnWidth
  else
    { otherwise create a new colinfo object, for this column }
    ColInfoStore.AddColInfo(index, index, ColumnWidth)
end;

function TColInfoStore.FindColInfo(index :integer) :TColInfo;
var
  loop :integer;
  tmpColInfo :TColInfo;
begin
  result := nil;

  for loop := 0 to Count - 1 do
  begin
    tmpColInfo := TColInfo(Items[loop]);
    if (tmpColInfo.Column = index) then
      result := tmpColInfo;
  end;
end;
{W8 END}

end.

